import platform
import time
from time import sleep
from typing import Dict, Optional, cast, Tuple

from rich.markdown import Markdown
from supabase import AsyncClient

from exegol.config.ConstantConfig import ConstantConfig
from exegol.config.EnvInfo import EnvInfo
from exegol.console.ConsoleFormat import get_display_date
from exegol.console.ExegolPrompt import ExegolRich
from exegol.console.TUI import ExegolTUI
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.exceptions.ExegolExceptions import CancelOperation
from exegol.manager.TaskManager import TaskManager
from exegol.model.LicensesTypes import LicensesEnumeration, LicenseType, LicenseEnrollment, EnrollmentForm
from exegol.utils.ExeLog import logger, console
from exegol.utils.LocalDatastore import LocalDatastore
from exegol.utils.MUID import MUID
from exegol.utils.SessionHandler import SessionHandler
from exegol.utils.SupabaseUtils import SupabaseUtils


class LicenseManager:
    __singleton_instance: Optional["LicenseManager"] = None

    def __init__(self) -> None:
        self.__session = SessionHandler()
        self.__user_session: Optional[AsyncClient] = None

    @classmethod
    async def get(cls) -> "LicenseManager":
        if cls.__singleton_instance is None:
            cls.__singleton_instance = cls()
            if not LocalDatastore().is_eula_accepted():
                await cls.__singleton_instance.eula_process()
                # For non-interactive commands skip first-time activation
                if not ParametersManager().accept_eula:
                    await cls.__singleton_instance.activate_exegol()
                else:
                    cls.__singleton_instance.__session.display_license()
            else:
                cls.__singleton_instance.__session.display_license()
        return cls.__singleton_instance

    async def eula_process(self) -> None:
        if not ParametersManager().accept_eula:
            while not await ExegolRich.Confirm("I confirm that I've read and accepted the EULA (https://docs.exegol.com/legal/eula)", default=False):
                await self.display_eula()
        LocalDatastore().update_eula(True)

    async def activate_exegol(self, skip_prompt: bool = False) -> None:
        # Do you want to activate your Exegol?
        if skip_prompt or await ExegolRich.Confirm("Do you want to activate your Exegol subscription now?", default=False):
            try:
                if await self.enroll_machine():
                    await self.__session.reload_session()
            except KeyboardInterrupt as e:
                if self.__user_session is not None:
                    await self.__user_session.auth.sign_out({"scope": "local"})
                raise e
        else:
            LocalDatastore().update_eula(False)
            logger.warning("[green]Community[/green] plan (i.e., free) is strictly limited to personal, non-commercial, educational, or research purposes.")
            sleep(1)
            if not await ExegolRich.Confirm("I confirm that I'm not trying to rip off the developers, who put their blood, sweat, and tears into Exegol, and that I would absolutely support the project if I were using it professionally. I'm not *that* kind of person.", default=False):
                logger.error("Wrong answer")
                while True:
                    logger.error("PUNK!")
                    sleep(1)
            logger.success("You can now use Exegol [green]Community[/green] Edition! Enjoy :)")
            LocalDatastore().update_eula(True)
        self.__session.display_license(as_info=True)

    async def revoke_exegol(self) -> None:
        if self.__session.is_enrolled():
            logger.warning("You are about to revoke your Exegol license. This action cannot be undone.")
            logger.warning("If you've recently enrolled this machine, you [red]won't[/red] be able to reuse this license immediately on another machine.")
            if await ExegolRich.Confirm("Are you sure you want to revoke your license locally?", default=False):
                await TaskManager.wait_for(TaskManager.TaskId.LoadLicense, clean_task=False)
                LocalDatastore().deactivate_license()
                logger.success("Exegol license successfully revoked.")
        else:
            logger.info("You don't have an active license to revoke.")

    def is_activated(self) -> bool:
        """Check if a valid license is available (license binding to an expired license while return False)"""
        return self.__session.get_license_type() != LicenseType.Community

    @staticmethod
    async def display_eula() -> None:
        if await ExegolRich.Confirm("Do you want to display EULA here?", default=True):
            # Display EULA
            with open(ConstantConfig.eula_path, 'r') as f:
                markdown = Markdown(f.read())
            console.print(markdown)

    async def enroll_machine(self) -> bool:
        # Login
        self.__user_session = await SupabaseUtils.login_user()
        # Enum license
        licenses_enum: Dict[str, LicensesEnumeration] = await SupabaseUtils.enum_licenses(self.__user_session.functions)

        if len(licenses_enum) == 0:
            logger.error("No active licenses assigned to this user are available.")
            return False
        # Select license variables
        selected_license_id: Optional[str] = None
        selected_license: Optional[Dict] = None
        revoke_previous_machine = False

        # Convert and process license ID and ISO datetime to printable date format
        table_data = {}
        id_match = {}
        i = 0
        for current_id, current_license in licenses_enum.items():
            i += 1
            current_license["valid_until"] = get_display_date(current_license["valid_until"])
            if current_license["last_seen"]:
                current_license["last_seen"] = get_display_date(current_license["last_seen"])
            table_data[str(i)] = current_license
            id_match[str(i)] = current_id

        while True:
            while selected_license_id is None:
                selected_license_id, selected_license = cast(Tuple[str, Dict],
                                                             await ExegolTUI.selectFromList(table_data, subject="a license ID", title="Select a license to enroll Exegol", default="1"))

                if selected_license.get("enrolled_on", ""):
                    logger.warning(f"The selected license is already used by another machine: [green]{selected_license.get('enrolled_on')}[/green]")
                    if not await ExegolRich.Confirm(f"Do you want to [red]revoke[/red] the license on [green]{selected_license.get('enrolled_on')}[/green] ?", default=True):
                        selected_license_id = None
                    else:
                        revoke_previous_machine = True
            if selected_license is None:
                logger.critical("No license was selected. Exiting.")
                raise CancelOperation
            try:
                # Check if the license type is known to the wrapper
                LicenseType[selected_license.get("type", "Unknown")]
            except KeyError as e:
                logger.critical("This license type is not supported the current version of the wrapper. Please update your wrapper first.")
                raise e
            try:
                current_os = EnvInfo.getHostOs().value.lower()
            except RuntimeError:
                # Fallback mthd if docker not yet init
                current_os = platform.system().lower()
                if current_os == "darwin":
                    current_os = "Mac"
            # Enroll
            enroll_form: EnrollmentForm = {
                "machine_id": MUID.get_current_muid(),
                "machine_name": platform.node(),
                "machine_os": current_os,
                "license_id": id_match[selected_license_id],
            }
            if revoke_previous_machine:
                enroll_form["revoke_previous_machine"] = True
            try:
                enrollment_response: LicenseEnrollment = await SupabaseUtils.activate_licenses(self.__user_session.functions, enroll_form)
                # Save license session
                if enrollment_response.get("next_token") is not None:
                    LocalDatastore().set(LocalDatastore.Key.TOKEN, enrollment_response.get("next_token"))
                    LocalDatastore().set(LocalDatastore.Key.SESSION, enrollment_response.get("session"))
                else:
                    raise NotImplementedError
                logger.success("Exegol successfully activated!")
                await self.__user_session.auth.sign_out({"scope": "local"})
                return True
            except CancelOperation:
                selected_license_id = None
                time.sleep(2)



from enum import Enum


class SyntaxFormat(Enum):
    port_sharing = "[default green not bold][<host_ipv4>:]<host_port>[-<end_host_port>][:<container_port>[-<end_container_port>]][:<protocol>][/default green not bold]"
    desktop_config = "[blue]proto[:ip[:port]][/blue]"
    volume = "/path/on/host/:/path/in/container/[blue][:ro|rw][/blue]"

    def __str__(self) -> str:
        return self.value



import json
import os
import platform
from enum import Enum
from json import JSONDecodeError
from pathlib import Path
from typing import Optional, List, Dict

from exegol.config.ConstantConfig import ConstantConfig
from exegol.utils.ExeLog import logger


class EnvInfo:
    """Class to identify the environment in which exegol runs to adapt
    the configurations, processes and messages for the user"""

    class HostOs(Enum):
        """Dictionary class for static OS Name"""
        WINDOWS = "Windows"
        LINUX = "Linux"
        MAC = "Mac"

    class DisplayServer(Enum):
        """Dictionary class for static Display Server"""
        WAYLAND = "Wayland"
        X11 = "X11"

    class DockerEngine(Enum):
        """Dictionary class for static Docker engine name"""
        WSL2 = "WSL2"
        HYPERV = "Hyper-V"
        DOCKER_DESKTOP = "Docker desktop"
        ORBSTACK = "Orbstack"
        LINUX = "Kernel"

    """Contain information about the environment (host, OS, platform, etc)"""
    # Shell env
    current_platform: str = "WSL" if "microsoft" in platform.release() else platform.system()  # Can be 'Windows', 'Linux' or 'WSL'
    is_linux_shell: bool = current_platform in ["WSL", "Linux"]
    is_windows_shell: bool = current_platform == "Windows"
    is_mac_shell: bool = not is_windows_shell and not is_linux_shell  # If not Linux nor Windows, its (probably) a mac
    __is_docker_desktop: bool = False
    __windows_release: Optional[str] = None
    # Host OS
    __docker_host_os: Optional[HostOs] = None
    __docker_engine: Optional[DockerEngine] = None
    # Docker desktop cache config
    __docker_desktop_resource_config: Optional[dict] = None
    # Architecture
    raw_arch = platform.machine().lower()
    arch = raw_arch
    if arch == "x86_64" or arch == "x86-64" or arch == "amd64":
        arch = "amd64"
    elif arch == "aarch64" or "armv8" in arch:
        arch = "arm64"
    elif "arm" in arch:
        if platform.architecture()[0] == '64bit':
            arch = "arm64"
        else:
            logger.error(f"Host architecture seems to be 32-bit ARM ({arch}), which is not supported yet. "
                         f"If possible, please install a 64-bit operating system (Exegol supports ARM64).")
        """
        if "v5" in arch:
            arch = "arm/v5"
        elif "v6" in arch:
            arch = "arm/v6"
        elif "v7" in arch:
            arch = "arm/v7"
        elif "v8" in arch:
            arch = "arm64"
        """
    else:
        logger.warning(f"Unknown / unsupported architecture: {arch}. Using 'AMD64' as default.")
        # Fallback to default AMD64 arch
        arch = "amd64"

    @classmethod
    def initData(cls, docker_info: Dict[str, str]) -> None:
        """Initialize information from Docker daemon data"""
        # Fetch data from Docker daemon
        docker_os = docker_info.get("OperatingSystem", "unknown").lower()
        docker_kernel = docker_info.get("KernelVersion", "unknown").lower()
        # Deduct a Windows Host from data
        cls.__is_docker_desktop = docker_os == "docker desktop"
        is_host_windows = cls.__is_docker_desktop and "microsoft" in docker_kernel
        is_orbstack = (docker_os == "orbstack" or "(containerized)" in docker_os) and "orbstack" in docker_kernel
        if is_host_windows:
            # Check docker engine with Windows host
            if "wsl2" in docker_kernel:
                cls.__docker_engine = cls.DockerEngine.WSL2
            else:
                cls.__docker_engine = cls.DockerEngine.HYPERV
            cls.__docker_host_os = cls.HostOs.WINDOWS
        elif cls.__is_docker_desktop:
            # If docker desktop is detected but not a Windows engine/kernel, it's (probably) a mac
            cls.__docker_engine = cls.DockerEngine.DOCKER_DESKTOP
            cls.__docker_host_os = cls.HostOs.MAC if cls.is_mac_shell else cls.HostOs.LINUX
        elif is_orbstack:
            # Orbstack is only available on Mac
            cls.__docker_engine = cls.DockerEngine.ORBSTACK
            cls.__docker_host_os = cls.HostOs.MAC
        else:
            # Every other case it's a linux distro and docker is powered from the kernel
            cls.__docker_engine = cls.DockerEngine.LINUX
            cls.__docker_host_os = cls.HostOs.LINUX

        if cls.__docker_engine == cls.DockerEngine.DOCKER_DESKTOP and cls.__docker_host_os == cls.HostOs.LINUX:
            logger.warning(f"Using Docker Desktop on Linux is not officially supported !")

    @classmethod
    def getHostOs(cls) -> HostOs:
        """Return Host OS
        Can be 'Windows', 'Mac' or 'Linux'"""
        # initData must be called from DockerUtils on client initialisation
        if cls.__docker_host_os is not None:
            return cls.__docker_host_os
        raise RuntimeError("Docker host OS is not initialized. Please call EnvInfo.initData() before using this method.")

    @classmethod
    def getDisplayServer(cls) -> DisplayServer:
        """Returns the display server
        Can be 'X11' or 'Wayland'"""
        session_type = os.getenv("XDG_SESSION_TYPE", "x11")
        if session_type == "wayland":
            return cls.DisplayServer.WAYLAND
        elif session_type in ["x11", "tty"]:  # When using SSH X11 forwarding, the session type is "tty" instead of the classic "x11"
            return cls.DisplayServer.X11
        else:
            # Should return an error
            logger.warning(f"Unknown session type {session_type}. Using X11 as fallback.")
            return cls.DisplayServer.X11

    @classmethod
    def getWindowsRelease(cls) -> str:
        # Cache check
        if cls.__windows_release is None:
            if cls.is_windows_shell:
                # From a Windows shell, python supply an approximate (close enough) version of windows
                cls.__windows_release = platform.win32_ver()[1]
            else:
                cls.__windows_release = "Unknown"
        return cls.__windows_release

    @classmethod
    def isWindowsHost(cls) -> bool:
        """Return true if Windows is detected on the host"""
        return cls.getHostOs() == cls.HostOs.WINDOWS

    @classmethod
    def isMacHost(cls) -> bool:
        """Return true if macOS is detected on the host"""
        return cls.getHostOs() == cls.HostOs.MAC

    @classmethod
    def isLinuxHost(cls) -> bool:
        """Return true if Linux is detected on the host"""
        return cls.getHostOs() == cls.HostOs.LINUX

    @classmethod
    def isWaylandAvailable(cls) -> bool:
        """Return true if wayland is detected on the host"""
        return cls.getDisplayServer() == cls.DisplayServer.WAYLAND or bool(os.getenv("WAYLAND_DISPLAY"))

    @classmethod
    def isDockerDesktop(cls) -> bool:
        """Return true if docker desktop is used on the host"""
        return cls.__is_docker_desktop

    @classmethod
    def isOrbstack(cls) -> bool:
        """Return true if docker desktop is used on the host"""
        return cls.__docker_engine == cls.DockerEngine.ORBSTACK

    @classmethod
    def getDockerEngine(cls) -> DockerEngine:
        """Return Docker engine type.
        Can be any of EnvInfo.DockerEngine"""
        # initData must be called from DockerUtils on client initialisation
        assert cls.__docker_engine is not None
        return cls.__docker_engine

    @classmethod
    def getShellType(cls) -> str:
        """Return the type of shell exegol is executed from"""
        if cls.is_linux_shell:
            return cls.HostOs.LINUX.value
        elif cls.is_windows_shell:
            return cls.HostOs.WINDOWS.value
        elif cls.is_mac_shell:
            return cls.HostOs.MAC.value
        else:
            return "Unknown"

    @classmethod
    def getDockerDesktopSettings(cls) -> Dict:
        """Applicable only for docker desktop on macos"""
        if cls.isDockerDesktop():
            if cls.__docker_desktop_resource_config is None:
                dir_path = None
                file_path = None
                if cls.is_mac_shell:
                    # Mac PATH
                    dir_path = ConstantConfig.docker_desktop_mac_config_path
                elif cls.is_windows_shell:
                    # Windows PATH
                    dir_path = ConstantConfig.docker_desktop_windows_config_path
                else:
                    # Windows PATH from WSL shell
                    # Find docker desktop config
                    config_file = list(Path("/mnt/c/Users").glob(f"*/{ConstantConfig.docker_desktop_windows_config_short_path}/settings-store.json"))
                    if len(config_file) == 0:
                        # Testing with legacy file name
                        config_file = list(Path("/mnt/c/Users").glob(f"*/{ConstantConfig.docker_desktop_windows_config_short_path}/settings.json"))
                        if len(config_file) == 0:
                            logger.warning(f"No docker desktop settings file found.")
                            return {}
                    file_path = config_file[0]
                if file_path is None:
                    assert dir_path is not None
                    # Try to find settings file with new filename or fallback to legacy filename for Docker Desktop older than 4.34
                    file_path = (dir_path / "settings-store.json") if (dir_path / "settings-store.json").is_file() else (dir_path / "settings.json")
                logger.debug(f"Loading Docker Desktop config from {file_path}")
                try:
                    with open(file_path, 'r') as docker_desktop_config:
                        cls.__docker_desktop_resource_config = json.load(docker_desktop_config)
                except FileNotFoundError:
                    logger.warning(f"Docker Desktop configuration file not found: '{file_path}'")
                    return {}
                except JSONDecodeError:
                    logger.critical(f"The Docker Desktop configuration file '{file_path}' is not a valid JSON. Please fix your configuration file first.")
            if cls.__docker_desktop_resource_config is None:
                logger.warning(f"Docker Desktop configuration couldn't be loaded.'")
            else:
                return cls.__docker_desktop_resource_config
        return {}

    @classmethod
    def getDockerDesktopResources(cls) -> List[str]:
        settings = cls.getDockerDesktopSettings()
        # Handle legacy settings key
        docker_desktop_resources = settings.get('FilesharingDirectories', settings.get('filesharingDirectories', []))
        logger.debug(f"Docker Desktop resources whitelist: {docker_desktop_resources}")
        return docker_desktop_resources

    @classmethod
    def isHostNetworkAvailable(cls) -> bool:
        if cls.isLinuxHost():
            return True
        elif cls.isOrbstack():
            return True
        elif cls.isDockerDesktop():
            settings = cls.getDockerDesktopSettings()
            # Handle legacy settings key
            res = settings.get('HostNetworkingEnabled', settings.get('hostNetworkingEnabled'))
            if res is None:
                logger.warning("Host network mode for Docker Desktop is not available, you need to upgrade Docker Desktop to enable it!")
            elif not res:
                logger.warning(
                    "Docker desktop now supports host network mode. However, this mode is currently [red]disabled[/red]. You need to manually change the configuration in your Docker Desktop settings to support host network sharing with Exegol.")
            if not res:
                logger.info("To share network ports (without host network) between the host and exegol, use the [bright_blue]--port[/bright_blue] parameter.")
                logger.verbose("Official doc: https://docs.docker.com/network/drivers/host/#docker-desktop")
            return res if res is not None else False
        logger.warning("Unknown or not supported environment for host network mode.")
        return False



import asyncio
import threading
from asyncio import Task
from enum import Enum
from threading import Thread
from typing import Coroutine, Dict, Optional, Any, Tuple, List, Set, cast

from exegol.utils.ExeLog import logger


class TaskManager:
    class TaskId(Enum):
        RemoteImageList = "remote_image_list"
        LocalImageList = "local_image_list"
        ImageList = "image_list"
        ContainerList = "container_list"
        LoadLicense = "load_license"
        # Threaded tasks
        RefreshSession = "session_refresh"  # Sub-task of LoadLicense

    __named_tasks: Dict[TaskId, Task] = {}
    __named_threads: Dict[TaskId, threading.Thread] = {}
    __tasks: Set[Task] = set()

    __waitable_tasks: Set[str] = {TaskId.LoadLicense.value}
    __threaded_tasks: Set[TaskId] = {TaskId.RefreshSession}

    @classmethod
    def add_task(cls, task_function: Coroutine, task_id: Optional[TaskId] = None) -> None:
        if task_id is not None:
            # Adding named tasks to the task manager
            if cls.__named_tasks.get(task_id) is None:
                logger.debug(f"Creating new {task_id} background task.")
                if task_id in cls.__threaded_tasks:
                    cls.__named_threads[task_id] = cls.__coro_to_thread(task_function, name=task_id.value)
                else:
                    cls.__named_tasks[task_id] = asyncio.create_task(task_function, name=task_id.value)
            else:
                logger.error(f"Task {task_id} already exists.")
        else:
            # Adding unnamed background tasks to the task manager
            new_task = asyncio.create_task(task_function)
            cls.__tasks.add(new_task)
            new_task.add_done_callback(cls.__tasks.discard)

    @classmethod
    async def wait_for(cls, task_id: TaskId, clean_task: bool = True) -> Any:
        if task_id in cls.__threaded_tasks:
            # Wait for the thread
            thread = cls.__named_threads.get(task_id)
            if thread is not None:
                await asyncio.to_thread(thread.join)
                if clean_task:
                    cls.__named_threads.pop(task_id)
        else:
            # Wait for the task
            task: Optional[Task] = cls.__named_tasks.get(task_id)
            if task is not None:
                await task
                if clean_task:
                    cls.__named_tasks.pop(task_id)
                return task.result()
        return None

    @classmethod
    async def gather(cls, *tasks: TaskId, clean_task: bool = True) -> Tuple[Any, ...]:
        all_tasks = []
        for t in tasks:
            current_task = cls.__named_tasks.get(t)
            if current_task is not None:
                all_tasks.append(current_task)
        results = await asyncio.gather(*all_tasks)
        if clean_task:
            for t in tasks:
                if t in cls.__named_tasks:
                    cls.__named_tasks.pop(t)
        return tuple(results)

    @staticmethod
    def __coro_to_thread(task: Coroutine, name: str) -> Thread:
        try:
            import asyncio
            import sys

            def run_in_thread(tasks: Coroutine) -> None:
                new_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(new_loop)
                try:
                    new_loop.run_until_complete(tasks)
                except (EOFError, KeyboardInterrupt, SystemExit) as err:
                    # Ignore exit from thread
                    logger.debug(f"Received exit signal from thread. Exiting thread: {type(err)} {err}")
                finally:
                    new_loop.close()
                    logger.debug("Thread closing")

            import threading
            license_thread = threading.Thread(target=run_in_thread, args=(task,), name=name, daemon=False)
            license_thread.start()

            return license_thread
        except Exception as e:
            logger.critical(f"Error during license tasks execution: {e}")
            raise

    @classmethod
    async def wait_for_all(cls, exit_mode: bool = False) -> None:
        tasks: List[Task] = []
        threads: List[Thread] = []
        for t in cls.__tasks.copy():
            if not t.done() and not t.cancelled():
                tasks.append(t)
        for t in cls.__named_tasks.values():
            if (t.get_name() in cls.__waitable_tasks and
                    not t.done() and not t.cancelled()):
                tasks.append(t)
            else:
                t.cancel()
        for th in cls.__named_threads.values():
            if th.is_alive():
                threads.append(th)

        if (len(tasks) + len(threads)) > 0:
            warning_sent = False
            logger.debug(f"Waiting for {len(tasks) + len(threads)} tasks")
            if len(threads) > 0:
                for th in threads:
                    if not th.is_alive():
                        continue
                    logger.debug(f"Waiting for {th.name} background task")
                    th.join(timeout=2 if exit_mode else 5)  # Max 10 seconds
                    if th.is_alive():
                        if not warning_sent:
                            logger.warning("Please wait for background tasks to complete.")
                            warning_sent = True
                        th.join(timeout=10)
            if len(tasks) > 0:
                logger.debug(f"Waiting for {', '.join([t.get_name() for t in tasks if not t.done()])} background task")
                # Task will not be awaited if the current context is canceled by user exit
                await asyncio.gather(*tasks)
            logger.debug(f"All background tasks done")

    @classmethod
    def sync_wait(cls, coroutine: Coroutine) -> Any:
        asyncio.run(coroutine)



from datetime import datetime
from multiprocessing import Queue
from typing import Optional, Tuple, Type, Union

import jwt

from exegol.config.ConstantConfig import ConstantConfig
from exegol.console.ExegolPrompt import ExegolRich
from exegol.exceptions.ExegolExceptions import CancelOperation, LicenseToleration, LicenseRevocation
from exegol.manager.TaskManager import TaskManager
from exegol.model.LicensesTypes import LicenseType
from exegol.utils.ExeLog import logger
from exegol.utils.KeyHandler import KeyHandler
from exegol.utils.LocalDatastore import LocalDatastore
from exegol.utils.MUID import MUID
from exegol.utils.MetaSingleton import MetaSingleton
from exegol.utils.SupabaseUtils import SupabaseUtils


class SessionHandler(metaclass=MetaSingleton):
    __PUB_KEY = """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA9rkavi/6ee+aDd+483oi
smKFmWYt2oQqPdxKDRblkKD8UORy2fXoCTQuW7LVyD4RsIAx4fibE+Zsj9Rg9rTY
S2HoSMUoJfSYktkpV3ZDD+Jm4rddG6bAeC2unFbNokmwOLz5hdjUgNwJ5fsyA61D
X1YYMIFH73WZS1zBNh0Q9OgqFkwSXz7PEygHea7RkQVsxAPULUCVpKSXLJWpIAh/
uIPcpd3EwkomPJKnjztydfF/iE4wx+54zdej1BReRd85QOfoRzAE+nS89kgjXiCi
xtgF1VPiq0g7RALUlZgox05XgmM1pvxaP6zEUkS5IPW/qv0cY/cC6DJRr9FDBrWx
owIDAQAB
-----END PUBLIC KEY-----"""
    __ALG = "ES256"

    def __init__(self) -> None:
        self.__key_handler = KeyHandler()
        # Raw JWT session
        self.__session: Optional[str] = None
        self.__token: Optional[str] = None
        # Session data
        self.__machine_id: Optional[str] = None
        self.__license: LicenseType = LicenseType.Community
        self.__license_id: Optional[str] = None
        self.__license_owner: Optional[str] = None
        self.__user_email: Optional[str] = None
        self.__expiration_date: Optional[datetime] = None
        self.__session_expiration_date: Optional[datetime] = None
        self.__session_issue_date: Optional[datetime] = None
        self.__is_enrolled: bool = False
        self.__current_muid = MUID.get_current_muid()
        # Fast loading local session
        self.__fast_load()
        # Local status
        self.__cert_force_refreshed = False
        self.__session_refreshed = False

    """
    Yes. This is where the license check is happening. We didn't mean this to be a difficult thing to bypass. Know one thing though.
    Not supporting Exegol today probably means future features won't ever see the light of day. But yes, you'll save on a few bucks tough.
    Open-source probably was a mistake then...
    """

    def is_enrolled(self) -> bool:
        """Return True if the wrapper is enrolled to an Exegol license expired or not"""
        return self.__is_enrolled

    def pro_feature_access(self) -> bool:
        return self.get_license_type().value >= LicenseType.Professional.value

    def enterprise_feature_access(self) -> bool:
        return self.get_license_type().value >= LicenseType.Enterprise.value

    def __is_session_valid(self) -> bool:
        if self.__license == LicenseType.Community:
            return True
        now = datetime.now()
        return (self.__session_expiration_date is not None and
                self.__expiration_date is not None and
                not (self.__session_expiration_date < now or
                     self.__expiration_date < now or
                     self.__machine_id != self.__current_muid))

    async def __refresh_thread_main(self, token: str, muid: str, return_queue: Queue) -> None:
        # Acquire refresh lock cross-process
        lock_path = ConstantConfig.exegol_config_path / ".session.lock"
        try:
            with open(lock_path, "x") as lock:
                lock.write(str(datetime.now().timestamp()))
            try:
                new_session = await self.__refresh_session(token, muid)
                return_queue.put((new_session, None))
            except Exception as e:
                return_queue.put((None, e))
            lock_path.unlink(missing_ok=True)
            return None
        except FileExistsError:
            logger.debug(f"Session refresh lock detected.")
            creation_time = None
            with open(lock_path, "r") as lock_file:
                try:
                    creation_time = float(lock_file.read())
                except ValueError:
                    pass
            if creation_time is None:
                # Fallback to file creation time
                try:
                    creation_time = lock_path.stat().st_birthtime  # type: ignore[attr-defined]
                except AttributeError:
                    creation_time = lock_path.lstat().st_ctime
            if creation_time is not None and creation_time > 0 and (datetime.now() - datetime.fromtimestamp(creation_time)).seconds >= 3600:
                logger.debug(f"Session refresh lock is older than 1 hour. Removing it.")
                lock_path.unlink()
                return await self.__refresh_thread_main(token, muid, return_queue)
        except PermissionError:
            logger.error(f"Permission denied in {ConstantConfig.exegol_config_path} directory. Exegol need Read/Write access to this directory.")
        return_queue.put((None, LicenseToleration))
        return None

    async def __refresh_session(self, token: str, muid: str) -> Optional[str]:
        logger.debug(f"Refreshing session")
        data = {"token": token, "machine_id": muid}
        if self.__machine_id is not None and self.__machine_id != muid:
            data["previous_id"] = self.__machine_id
        # Token comparison check in case lock have been overwritten
        _, db_token = LocalDatastore().get_license()
        if token != db_token:
            logger.debug(f"Licence token mismatch detected. Refresh session cancelled.")
            raise LicenseToleration
        # Rotate token
        try:
            refresh_token = await SupabaseUtils.rotate_token(data)
        except LicenseRevocation as e:
            LocalDatastore().deactivate_license()
            raise e
        # Save refresh token
        if refresh_token is not None:
            logger.debug(f"Storing refresh token response")
            LocalDatastore().set(LocalDatastore.Key.TOKEN, refresh_token)

        # Get session
        try:
            refresh_session = await SupabaseUtils.refresh_session({"token": refresh_token})
        except LicenseRevocation as e:
            LocalDatastore().deactivate_license()
            raise e
        # Save refresh token
        if refresh_session is not None:
            logger.debug(f"Storing session response")
            LocalDatastore().set(LocalDatastore.Key.SESSION, refresh_session)

        logger.debug(f"Session refreshed and stored")
        return refresh_session

    def __session_parsing(self) -> None:
        """
        Parse JWT session and extract data
        Raise many exceptions if the session is invalid
        :return:
        """
        if self.__session is None:
            raise jwt.InvalidTokenError
        session = jwt.decode(self.__session,
                             self.__key_handler.getKey(),
                             algorithms=[self.__ALG],
                             options={"require": ["iat", "iss", "aud", "exp"], "verify_exp": False, "verify_iat": False},
                             audience="urn:exegol:wrapper",
                             issuer=self.__key_handler.getSubject(), )
        logger.debug(f"Local session: {session}")
        self.__machine_id = session["machine_id"]
        self.__license_id = session["license_id"]
        self.__license_owner = session["license_owner"]
        self.__user_email = session["username"]
        self.__expiration_date = datetime.fromtimestamp(session["expiration_date"])
        self.__session_expiration_date = datetime.fromtimestamp(session["exp"])
        self.__session_issue_date = datetime.fromtimestamp(session["iat"])
        self.__license = LicenseType[session.get("license", "Community")]

    def __fast_load(self) -> bool:
        """
        Sync fast load from init for Completer usage / UserConfig generation etc
        :return:
        """
        logger.debug("Fast loading session from datastore")
        if self.__token is None:
            self.__session, self.__token = LocalDatastore().get_license()
        self.__is_enrolled = self.__token is not None
        if self.__token is None:
            return False
        now = datetime.now()
        # Verify JWT
        try:
            if not self.__session:
                # If fast_load is True, but we only have a refresh token and no session
                return False
            self.__session_parsing()

            if self.__session_expiration_date is None or self.__expiration_date is None:
                raise LicenseToleration

            # Manually handle JWT expiration after payload parsing
            if self.__session_expiration_date < now:
                raise LicenseToleration

            # - Check local machine ID
            # - Expiration date still valid
            if self.__machine_id != self.__current_muid or self.__expiration_date < now:
                raise LicenseToleration
            logger.debug("Session is valid and fully processed")
            return True
        except KeyError:
            logger.error("Your version of Exegol wrapper doesn't support this license type. Please update your exegol wrapper.")
            self.__license = LicenseType.Community
        except jwt.ExpiredSignatureError:
            pass
        except jwt.InvalidTokenError:
            pass
        except CancelOperation:
            pass
        except LicenseToleration:
            # - Signed by JWT Cert keymax 7j (en tant qu'offline)
            # - Warning after 24h (post-expiration) offline
            # - Interactive warning message after 3j+
            logger.debug("License toleration detected")
            if self.__license != LicenseType.Community:
                if (self.__expiration_date is not None and
                        self.__session_expiration_date is not None and
                        self.__expiration_date > now):
                    expiration_delta = now - self.__session_expiration_date
                    # After 7 days the grace period expired
                    if expiration_delta.days >= 7:
                        self.__license = LicenseType.Community
                else:
                    self.__license = LicenseType.Community
        return False

    async def reload_session(self, force_refresh: bool = False) -> bool:
        """

        :return: True is the session was reloaded
        """
        logger.debug("Fetch session from datastore")
        if self.__token is None:
            self.__session, self.__token = LocalDatastore().get_license()
        self.__is_enrolled = self.__token is not None
        if self.__token is None:
            return False
        now = datetime.now()
        # Verify JWT
        try:
            await self.__key_handler.refresh_certificate()  # Refresh cert if not yet ready
            # Check when session need to be reloaded from server (because session is missing but token is here, or because it's time to refresh)
            if not self.__session or force_refresh:
                if self.__session_issue_date and (now - self.__session_issue_date).seconds <= 300:
                    # Too early to use the token
                    raise LicenseToleration
                refresh_queue: Queue[Tuple[Optional[str], Optional[Union[Exception, Type[Exception]]]]] = Queue()
                # Critical task in thread to prevent abort and loose session binding
                TaskManager.add_task(self.__refresh_thread_main(self.__token, self.__current_muid, refresh_queue),
                                     TaskManager.TaskId.RefreshSession)
                await TaskManager.wait_for(TaskManager.TaskId.RefreshSession)
                refresh_error: Optional[Union[Exception, Type[Exception]]]
                # Fetch Result of refresh thread
                self.__session, refresh_error = refresh_queue.get()
                self.__token = None
                refresh_queue.close()
                self.__session_refreshed = True
                if refresh_error is not None:
                    raise refresh_error
                elif self.__session is None:
                    raise CancelOperation

            elif not self.__session:
                # If fast_load is True, but we only have a refresh token and no session
                return False
            self.__session_parsing()

            if self.__session_expiration_date is None or self.__expiration_date is None:
                raise jwt.InvalidTokenError

            # Manually handle JWT expiration after payload parsing
            if self.__session_expiration_date < now:
                raise jwt.ExpiredSignatureError

            # - Check local machine ID
            # - Expiration date still valid
            if self.__machine_id != self.__current_muid or self.__expiration_date < now:
                if (now - self.__session_expiration_date).seconds <= 300:
                    raise LicenseToleration
                if not force_refresh and not self.__session_refreshed:
                    return await self.reload_session(force_refresh=True)
                else:
                    raise CancelOperation
            logger.debug("Session is valid and fully processed")
            return True
        except KeyError:
            logger.error("Your version of Exegol wrapper doesn't support this license type. Please update your exegol wrapper.")
            self.__license = LicenseType.Community
        except jwt.ExpiredSignatureError:
            if force_refresh or self.__session_refreshed:
                logger.critical("Your system doesn't seem to be on time. Please check your system settings.")
            logger.debug("Session expired, reloading it")
            return await self.reload_session(force_refresh=True)
        except jwt.InvalidTokenError as e:
            logger.debug(f"Invalid token error: {e}")
            if force_refresh or self.__session_refreshed:
                if not self.__cert_force_refreshed:
                    cert_changed = await self.__key_handler.refresh_certificate(force_refresh=True)
                    self.__cert_force_refreshed = True
                    if cert_changed:
                        return await self.reload_session()
                self.__license = LicenseType.Community
                logger.error("An error occurred, your license cannot be processed. Please update your wrapper to the latest version or contact support.")
            return await self.reload_session(force_refresh=True)
        except CancelOperation:
            self.__license = LicenseType.Community
            await self.__license_expired(True)
        except LicenseRevocation:
            self.__license = LicenseType.Community
            await ExegolRich.Acknowledge("Your license has been revoked. Re-activation is required.")
        except LicenseToleration:
            # - Signed by JWT Cert keymax 7j (en tant qu'offline)
            # - Warning after 24h (post-expiration) offline
            # - Interactive warning message after 3j+
            logger.debug("License toleration detected")
            if self.__license != LicenseType.Community:
                if (self.__expiration_date is not None and
                        self.__session_expiration_date is not None and
                        self.__expiration_date > now):
                    expiration_delta = now - self.__session_expiration_date
                    # After 7 days the grace period expired
                    if expiration_delta.days < 7:
                        days_remaining = 7 - expiration_delta.days
                        # Warning after 2d offline
                        if expiration_delta.days >= 2:
                            logger.warning(f"Offline use of your {self.get_license_type_display()} features is allowed for up to 7 days ({expiration_delta.days}/7).")
                            # More warning message after 5j
                            if expiration_delta.days == 5:
                                logger.warning(f"{self.get_license_type_display()} features will be deactivated in {days_remaining} days if Exegol can't connect to the Internet before then.")
                            else:
                                await ExegolRich.Acknowledge(f"{self.get_license_type_display()} features will be deactivated in {days_remaining} days if Exegol can't connect to the Internet before then.")
                    else:
                        self.__license = LicenseType.Community
                        await ExegolRich.Acknowledge("Offline usage period exceeded. Connect Exegol back to the internet to re-enable Pro features.")
                else:
                    self.__license = LicenseType.Community
                    await self.__license_expired(False)
        return False

    @staticmethod
    async def __license_expired(have_internet: bool) -> None:
        await ExegolRich.Acknowledge(f"License has expired. Pro features not available anymore. Renew subscription{'' if have_internet else ' and connect to the Internet.'} to re-enable.")
        if await ExegolRich.Confirm("Do you want to deactivate your wrapper and cancel license auto-refresh?", default=False):
            LocalDatastore().deactivate_license()
        else:
            logger.info("Keeping local license binding. Exegol needs Internet access to enable Pro features.")

    def display_license(self, as_info: bool = False) -> None:
        display = f"Exegol "
        if self.__license is LicenseType.Community:
            display += f"[green]{self.__license.name}[/green] (personal use only)"
        elif self.__license is LicenseType.Professional:
            display += f"[gold3]{self.__license.name}[/gold3] licensed to [green]{self.__user_email}[/green]"
        elif self.__license is LicenseType.Enterprise:
            display += f"[gold3]{self.__license.name}[/gold3] licensed to [green]{self.__license_owner}[/green] / [green]{self.__user_email}[/green]"
        else:
            raise NotImplementedError

        if self.__license is not LicenseType.Community:
            logger.debug(f"Using license {self.__license_id} from machine {self.__machine_id}")
        if as_info:
            logger.info(display)
        else:
            logger.verbose(display)
        if self.__expiration_date is not None:
            logger.verbose(f"License valid until {self.__expiration_date}")

    def get_license_type_display(self) -> str:
        if self.__license is not None:
            if self.__license != LicenseType.Community:
                return f"[gold3]{self.__license.name}[/gold3]"
        return f"[green]{LicenseType.Community.name}[/green]"

    def get_license_type(self) -> LicenseType:
        if self.__license is not None:
            return self.__license
        return LicenseType.Community

    async def get_registry_auth(self, registry: str, tag: str) -> dict:
        await TaskManager.wait_for(TaskManager.TaskId.LoadLicense, clean_task=False)
        # Check if session still valid
        if not self.__is_session_valid():
            # Reload session if needed
            if not await self.reload_session(force_refresh=True) or not self.__is_session_valid():
                logger.critical("You cannot access the official registry without access to Exegol license servers.")
        if self.__license == LicenseType.Community or self.__session is None:
            logger.critical("Pro/Enterprise license required to download non-Free images.")
            raise SystemExit(1)
        # Get registry token for specific registry
        return {"username": "__token__", "password": await SupabaseUtils.registry_access({"image_name": registry, "tag": tag}, self.__session)}



from rich.status import Status

from exegol.utils.ExeLog import console, ConsoleLock


class ExegolStatus(Status):

    def __init__(self, status: str, **kwargs) -> None:
        super().__init__(status, console=console, **kwargs)

    async def __aenter__(self) -> "ExegolStatus":
        await ConsoleLock.acquire()
        try:
            self.__enter__()
            return self
        except Exception as e:
            ConsoleLock.release()
            raise e

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        try:
            self.__exit__(exc_type, exc_val, exc_tb)
        finally:
            ConsoleLock.release()



import json
import os
import sys
from json import JSONEncoder, JSONDecodeError
from pathlib import Path
from types import MethodType, FunctionType
from typing import Union, Dict, cast, Optional, Set, Any, List

import yaml
import yaml.parser

from exegol.config.ConstantConfig import ConstantConfig
from exegol.utils.ExeLog import logger
from exegol.utils.FsUtils import mkdir, get_user_id


class DataFileUtils:

    class ObjectJSONEncoder(JSONEncoder):
        def default(self, o: object) -> Dict:
            result = {}
            for key, value in o.__dict__.items():
                if key.startswith("_"):
                    continue
                result[key] = value
            return result

    def __init__(self, file_path: Union[Path, str], file_type: str, dynamic_default: Optional[Dict] = None):
        """Generic class for datastorage in config file.
        :param file_path: can be a Path object or a string. If a string is supplied it will be automatically place inside the default exegol directory.
        :param file_type: defined the type of file to create. It can be 'yml' or 'json'.
        :param dynamic_default: An optional dictionnary that can be used to dynamically assign a default value to a config key. The key is the config name and the value is a class method or function.
        """
        if type(file_path) is str:
            file_path = ConstantConfig.exegol_config_path / file_path
        if file_type not in ["yml", "yaml", "json"]:
            raise NotImplementedError(f"The file type '{file_type}' is not implemented")
        # Config file options
        self._file_path: Path = cast(Path, file_path)
        self.__file_type: str = file_type
        self.__config_upgrade: bool = False

        # Dynamic default
        self.__dynamic_default: Dict[str, Union[MethodType, FunctionType]] = dynamic_default if dynamic_default is not None else dict()

        self._raw_data: Any = None

        # Process
        self.__load_file()

    def _get_dynamic_default(self, attr_name: str) -> Union[str, int, bool]:
        if attr_name in self.__dynamic_default.keys():
            return self.__dynamic_default[attr_name]()
        raise NotImplementedError(f"The dynamic default value is not define for attribute '{attr_name}'")

    def __load_file(self) -> None:
        """
        Function to load the file and the corresponding parameters
        :return:
        """
        if not self._file_path.parent.is_dir():
            logger.verbose(f"Creating config folder: {self._file_path.parent}")
            mkdir(self._file_path.parent)
        if not self._file_path.is_file():
            logger.verbose(f"Creating default file: {self._file_path}")
            self._create_config_file()
        else:
            self._parse_config()
            if self.__config_upgrade:
                logger.verbose("Upgrading config file")
                self._create_config_file()

    def _build_file_content(self) -> str:
        """
        This function build the default file content. Called when the file doesn't exist yet or have been upgrade and need to be updated.
        :return:
        """
        raise NotImplementedError(f"The '_build_default_file' method hasn't been implemented in the '{self.__class__}' class.")

    def _create_config_file(self) -> None:
        """
        Create or overwrite the file content to the default / current value depending on the '_build_default_file' that must be redefined in child class.
        :return:
        """
        try:
            with open(self._file_path, 'w') as file:
                file.write(self._build_file_content())
            if sys.platform == "linux" and os.getuid() == 0:
                user_uid, user_gid = get_user_id()
                os.chown(self._file_path, user_uid, user_gid)
        except PermissionError as e:
            logger.critical(f"Unable to open the file '{self._file_path}' ({e}). Please fix your file permissions or run exegol with the correct rights.")
        except OSError as e:
            logger.critical(f"A critical error occurred while interacting with filesystem: [{type(e)}] {e}")

    def _parse_config(self) -> None:
        data: Dict = {}
        with open(self._file_path, 'r') as file:
            try:
                if self.__file_type == "yml":
                    data = yaml.safe_load(file)
                elif self.__file_type == "json":
                    data = json.load(file)
            except yaml.parser.ParserError:
                logger.error("Error while parsing exegol config file ! Check for syntax error.")
            except JSONDecodeError:
                logger.error(f"Error while parsing exegol data file {self._file_path} ! Check for syntax error.")
        if data is None:
            logger.warning(f"Exegol was unable to load the file {self._file_path}. Restoring it to its original state.")
            self._create_config_file()
        else:
            self._raw_data = data
            self._process_data()

    def __load_config(self, data: dict, config_name: str, default: Optional[Union[bool, str, int, List[str]]],
                      choices: Optional[Set[str]] = None) -> Union[bool, str, int, List[str]]:
        """
        Function to automatically load a data from a dict object. This function can handle limited choices and default value.
        If the parameter don't exist,a reset flag will be raised.
        :param data: Dict data to retrieve the value from
        :param config_name: Key name of the config to find
        :param default: Default value is the value hasn't been set yet
        :param choices: (Optional) A limit set of acceptable values
        :return: This function return the value for the corresponding config_name.
        """
        try:
            if config_name not in data.keys():
                logger.debug(f"Config {config_name} has not been found in Exegol '{self._file_path.name}' config file. The file will be upgrade.")
                self.__config_upgrade = True
                return default if default is not None else self._get_dynamic_default(config_name)
            result = data.get(config_name)
            if result is None:
                return default if default is not None else self._get_dynamic_default(config_name)
            elif choices is not None and result not in choices:
                default = default if default is not None else self._get_dynamic_default(config_name)
                logger.warning(f"The configuration is incorrect! "
                               f"The user has configured the '{config_name}' parameter with the value '{result}' "
                               f"which is not one of the allowed options ({', '.join(choices)}). Using default value: {default}.")
                return default
            return result
        except TypeError:
            logger.error(f"Error while loading {config_name}! Using default config.")
        return default if default is not None else self._get_dynamic_default(config_name)

    def _load_config_bool(self, data: dict, config_name: str, default: Optional[bool] = None,
                          choices: Optional[Set[str]] = None) -> bool:
        """
        Function to automatically load a BOOL from a dict object. This function can handle limited choices and default value.
        If the parameter don't exist,a reset flag will be raised.
        :param data: Dict data to retrieve the value from
        :param config_name: Key name of the config to find
        :param default: Default value is the value hasn't been set yet
        :param choices: (Optional) A limit set of acceptable values
        :return: This function return the value for the corresponding config_name
        """
        return cast(bool, self.__load_config(data, config_name, default, choices))

    def _load_config_str(self, data: dict, config_name: str, default: Optional[str] = None, choices: Optional[Set[str]] = None) -> str:
        """
        Function to automatically load a STR from a dict object. This function can handle limited choices and default value.
        If the parameter don't exist,a reset flag will be raised.
        :param data: Dict data to retrieve the value from
        :param config_name: Key name of the config to find
        :param default: Default value is the value hasn't been set yet
        :param choices: (Optional) A limit set of acceptable values
        :return: This function return the value for the corresponding config_name
        """
        return cast(str, self.__load_config(data, config_name, default, choices))

    def _load_config_int(self, data: dict, config_name: str, default: Optional[int] = None, choices: Optional[Set[str]] = None) -> int:
        """
        Function to automatically load a INT from a dict object. This function can handle limited choices and default value.
        If the parameter don't exist,a reset flag will be raised.
        :param data: Dict data to retrieve the value from
        :param config_name: Key name of the config to find
        :param default: Default value is the value hasn't been set yet
        :param choices: (Optional) A limit set of acceptable values
        :return: This function return the value for the corresponding config_name
        """
        config = cast(Union[str, int], self.__load_config(data, config_name, default, choices))
        try:
            return int(config)
        except ValueError:
            logger.critical(f"Invalid value for {config_name}: received '{config}' instead of a number. Please use a correct format.")
            exit(1)

    def _load_config_path(self, data: dict, config_name: str, default: Optional[Path] = None) -> Path:
        """
        Function to automatically load a PATH from a dict object. This function can handle limited choices and default value.
        If the parameter don't exist,a reset flag will be raised.
        :param data: Dict data to retrieve the value from
        :param config_name: Key name of the config to find
        :param default: Default value is the value hasn't been set yet
        :return: This function return the value for the corresponding config_name
        """
        try:
            result = data.get(config_name)
            if result is None:
                logger.debug(f"Config {config_name} has not been found in Exegol '{self._file_path.name}' config file. The file will be upgrade.")
                self.__config_upgrade = True
                return default if default is not None else cast(Path, self._get_dynamic_default(config_name))
            return Path(result).expanduser()
        except TypeError:
            logger.error(f"Error while loading {config_name}! Using default config.")
        return default if default is not None else cast(Path, self._get_dynamic_default(config_name))

    def _load_config_list_str(self, data: dict, config_name: str, default: Optional[List[str]] = None) -> List[str]:
        """
        Function to automatically load a STR from a dict object. This function can handle limited choices and default value.
        If the parameter don't exist,a reset flag will be raised.
        :param data: Dict data to retrieve the value from
        :param config_name: Key name of the config to find
        :return: This function return the value for the corresponding config_name
        """
        return cast(List[str], self.__load_config(data, config_name, default=default if default is not None else list(), choices=None))

    def _process_data(self) -> None:
        raise NotImplementedError(f"The '_process_data' method hasn't been implemented in the '{self.__class__}' class.")



import logging
from enum import Enum
from typing import Optional, Union, List, Tuple, Dict, cast

from gotrue.errors import AuthApiError, AuthRetryableError, AuthInvalidCredentialsError, AuthUnknownError
from httpx import ConnectError, TransportError
from postgrest import APIError, AsyncFilterRequestBuilder, \
    AsyncMaybeSingleRequestBuilder, AsyncSingleRequestBuilder, AsyncSelectRequestBuilder, APIResponse
from supabase import create_async_client, AsyncClient
from supabase.lib.client_options import AsyncClientOptions
from supabase_functions import AsyncFunctionsClient
from supabase_functions.errors import FunctionsHttpError, FunctionsRelayError

from exegol.config.ConstantConfig import ConstantConfig
from exegol.console.ExegolPrompt import ExegolRich
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.exceptions.ExegolExceptions import CancelOperation, LicenseToleration, LicenseRevocation
from exegol.model.LicensesTypes import LicenseSession, TokenRotate, LicenseEnrollment, EnrollmentForm
from exegol.model.SupabaseModels import SupabaseImage
from exegol.utils.ExeLog import logger


class SupabaseUtils:
    class LicenseAction(Enum):
        LicenseEnum = "license_enumeration"
        LicenseActivation = "license_activation"
        TokenRotate = "token_rotate"
        TokenSession = "token_session"

        RegistryAccess = "registry_access"

        GetCertificate = "machine_certificate"

    @classmethod
    async def __create_client(cls) -> AsyncClient:
        return await create_async_client(ConstantConfig.SUPABASE_URL, ConstantConfig.SUPABASE_KEY, options=AsyncClientOptions(
            persist_session=False,
            postgrest_client_timeout=10,
            schema="public",
        ))

    @classmethod
    async def login_user(cls) -> AsyncClient:
        if ParametersManager().offline_mode:
            logger.critical("Exegol can't be activated without Internet access.")
        # Login
        supabase_client = await cls.__create_client()
        logger.info("Authenticating to Exegol")
        auth_response = None
        email = None
        for i in range(3):
            if email is None:
                email = await ExegolRich.Ask("Enter your exegol email account")
            if email.strip() == "":
                email = None
                logger.error("You must supply a valid email account, please retry")
                continue
            try:
                logger.success("To link your account to the wrapper, you need to generate a login code "
                               "from your dashboard: https://dashboard.exegol.com/otp")
                token = await ExegolRich.Ask("Enter your login token")
                auth_response = await supabase_client.auth.verify_otp({
                    "type": "magiclink",
                    "email": email,
                    "token": token,
                })
                break
            except AuthInvalidCredentialsError:
                logger.error("You must supply a valid email account, please retry")
            except AuthApiError as e:
                if "Invalid email" in e.message:
                    email = None
                logger.error(f"{e}, please retry")
            except AuthRetryableError as e:
                if e.status == 503:
                    logger.critical("The Exegol servers are currently unavailable, please try again later.")
                elif e.status == 0 and e.code is None:
                    logger.critical("You can't activate Exegol without internet access.")
                else:
                    logger.debug(f"Received error [{e.status}] {e.code} {e.message}")
                    logger.error(f"Unable to authenticate you to Exegol, retry later. {e.message}")
            except AuthUnknownError as e:
                logger.error(f"An unknown error occurred during authentication, retry later. {e.message}")
        if auth_response is None or auth_response.user is None:
            logger.critical("Unable to authenticate, aborting.")
        logger.success("Successfully authenticated")
        return supabase_client

    @classmethod
    async def __call_licenses_endpoint(cls, supabase_client: AsyncFunctionsClient, function_name: LicenseAction,
                                       body: Optional[Dict] = None, auth_token: Optional[str] = None) -> Dict:
        """
        This function raises the following exception that should be caught accordingly:
            - FunctionsRelayError (supabase error)
            - ConnectError (no internet)
            - TransportError (function unavailable)
            - FunctionsHttpError (app error)
        :param supabase_client: Supabase client instance (auth or anonymous)
        :param function_name: Function name to call (enum, activate, renew, registry_access, ...)
        :param body: Form (if any)
        :param auth_token: Session token (if any)
        :return:
        """
        if ParametersManager().offline_mode:
            # This check must be made before
            raise NotImplementedError
        headers: Dict[str, str] = {"app-action": function_name.value}
        options: Dict[str, Union[str, Dict]] = {"headers": headers}
        if function_name != cls.LicenseAction.GetCertificate:
            options["responseType"] = "json"
        if body:
            options["body"] = body
        if auth_token:
            headers["app-session"] = auth_token
        result = await supabase_client.invoke("licenses-endpoint", invoke_options=options)
        if type(result) is bytes:
            return {"result": result}
        return result

    @classmethod
    async def __execute(cls, supabase_query: Union[AsyncFilterRequestBuilder, AsyncMaybeSingleRequestBuilder, AsyncSingleRequestBuilder, AsyncSelectRequestBuilder]) -> Optional[APIResponse]:
        if ParametersManager().offline_mode:
            # This check must be made before
            raise NotImplementedError
        try:
            return await supabase_query.execute()
        except ConnectError as e:
            logger.debug(f"Supabase connect error: {e}")
            if "CERTIFICATE_VERIFY_FAILED" in str(e):
                logger.warning("TLS certificate verification failed while contacting Exegol servers. Do you have a proxy server ? If so, python probably doesn't trust your private CA.")
                logger.warning("Check the Exegol documentation to fix the issue: https://docs.exegol.com/troubleshooting#tls-certificate-verification-issues")
            else:
                logger.warning("No internet connection available. Skipping online queries.")
        except TransportError:
            logger.warning("An error occurred while contacting the server.")
        except APIError as e:
            if e.code == 503:
                logger.warning("The Exegol servers are currently unavailable, image status cannot be retrieved.")
            else:
                logger.debug(e.details)
                logger.error(f"[{e.code}] An error occurred while contacting Exegol servers: {e.message}")
                logger.warning("Some information might be missing.")
        return None

    ### License calls ###

    @classmethod
    async def get_cert(cls) -> bytes:
        """
        Get new public certificate from license server.
        :return: Public certificate as string.
        """
        if ParametersManager().offline_mode:
            logger.critical("You need internet access to validate Exegol license.")
        try:
            data: Dict[str, bytes] = await cls.__call_licenses_endpoint((await cls.__create_client()).functions, cls.LicenseAction.GetCertificate)
            cert = data.get("result")
            if cert is None:
                raise FunctionsRelayError
            return cert
        except ConnectError as e:
            logger.critical("Exegol license server can't be reached without Internet access")
            raise e
        except (FunctionsRelayError, TransportError) as e:
            logger.critical("Exegol license server seems to be unavailable for now. Please retry later.")
            raise e
        except FunctionsHttpError as e:
            if e.status in [500, 503, 504, 546]:
                logger.critical("The Exegol servers are currently unavailable, please try again later.")
            else:
                logger.critical(f"Unable to enumerate licenses from exegol servers: [{e.status}] {e.message}")
            raise e

    @classmethod
    async def enum_licenses(cls, supabase_client: AsyncFunctionsClient) -> Dict:
        """
        Enumerate all licenses available for the current user.
        :param supabase_client:
        :return:
        """
        if ParametersManager().offline_mode:
            logger.critical("Exegol can't be activated without Internet access.")
        try:
            return await cls.__call_licenses_endpoint(supabase_client, cls.LicenseAction.LicenseEnum)
        except ConnectError as e:
            logger.critical("Exegol license server can't be reached without Internet access.")
            raise e
        except (FunctionsRelayError, TransportError) as e:
            logger.critical("Exegol license server seems to be unavailable for now. Please retry later.")
            raise e
        except FunctionsHttpError as e:
            if e.status in [403, 503, 504, 546]:
                logger.critical("The Exegol servers are currently unavailable, please try again later.")
            else:
                logger.critical(f"Unable to enumerate licenses from exegol servers: [{e.status}] {e.message}")
            raise e

    @classmethod
    async def activate_licenses(cls, supabase_client: AsyncFunctionsClient, form: EnrollmentForm) -> LicenseEnrollment:
        if ParametersManager().offline_mode:
            logger.critical("You can't activate Exegol without Internet access.")
        try:
            return cast(LicenseEnrollment, await cls.__call_licenses_endpoint(supabase_client, cls.LicenseAction.LicenseActivation, cast(dict, form)))
        except ConnectError:
            logger.error("Exegol license server can't be reached without Internet access.")
            raise CancelOperation
        except (FunctionsRelayError, TransportError):
            logger.error("Exegol license server seems to be unavailable for now. Please retry later.")
            raise CancelOperation
        except FunctionsHttpError as e:
            if e.status in [401, 404, 406]:
                logger.raw(msg=f"[bold red][-][/bold red] {e.message}", markup=True, emoji=True, level=logging.ERROR)
                logger.empty_line(log_level=logging.ERROR)
            elif e.status == 400:
                if e.message == "Bad Request":
                    logger.error("Exegol license server seems to be unavailable for now. Check if your exegol is up-to-date and retry.")
                else:
                    logger.error(f"An error occurred when activating Exegol: {e.message}")
            elif e.status in [403, 503, 504, 546]:
                logger.error(f"The license server is currently unavailable. Please try again later.")
                logger.debug(f"Error code [{e.status}] {e}")
            elif e.status >= 500:
                logger.error(f"An error has occurred in the license server. You can contact support to help you. Error: [{e.status}] {e.message}")
            else:
                logger.error(f"An unknown error occurred when activating Exegol: {e.message} {e}")
            raise CancelOperation

    @staticmethod
    def __handle_session_error(e: FunctionsHttpError):
        if e.status == 403:
            # 403 invalid token (revoked)
            logger.error("Your license has been revoked on this machine.")
            raise LicenseRevocation
        elif e.status == 410:
            # 410 invalid licenses (expired)
            logger.error("Your license has expired. Please re-activate your exegol with another license.")
            raise LicenseRevocation
        elif e.status == 425:
            # 425 sessions cannot be renewed yet
            logger.debug("Session refresh rate limited.")
        elif e.status == 400:
            logger.warning("Received an unexpected error from license server. Check if your wrapper is up-to-date.")
        elif e.status == 500:
            logger.verbose(f"The Exegol servers are currently unavailable, please try again later.")
        elif e.status in [503, 504, 546]:
            logger.debug(f"License server seems to be unavailable for now: code {e.status} {e.message}. Retry later.")
        else:
            logger.error(f"An unknown error occurred when renewing session: [{e.status}] {e.message}")

    @classmethod
    async def rotate_token(cls, form: dict) -> str:
        if ParametersManager().offline_mode:
            raise LicenseToleration
        try:
            result = cast(TokenRotate, await cls.__call_licenses_endpoint((await cls.__create_client()).functions,
                                                                          cls.LicenseAction.TokenRotate,
                                                                          form))
            token = result.get("next_token")
            if token is None:
                raise FunctionsRelayError("Received an empty response from license server.")
            return token
        except ConnectError:
            logger.error("Exegol license server can't be reached without Internet access")
            raise LicenseToleration
        except (FunctionsRelayError, TransportError) as e:
            # Error during http request
            logger.debug(f"Error during session refresh. Network error, retry later: {e}")
            raise LicenseToleration
        except FunctionsHttpError as e:
            cls.__handle_session_error(e)
            raise LicenseToleration

    @classmethod
    async def refresh_session(cls, form: dict) -> Optional[str]:
        if ParametersManager().offline_mode:
            raise LicenseToleration
        try:
            result = cast(LicenseSession, await cls.__call_licenses_endpoint((await cls.__create_client()).functions,
                                                                             cls.LicenseAction.TokenSession,
                                                                             form))
            return result.get("session")
        except ConnectError:
            logger.error("Exegol license server can't be reached without Internet access")
            raise LicenseToleration
        except (FunctionsRelayError, TransportError) as e:
            # Error during http request
            logger.debug(f"Error during session refresh. Network error, retry later: {e}")
            raise LicenseToleration
        except FunctionsHttpError as e:
            cls.__handle_session_error(e)
            raise LicenseToleration

    @classmethod
    async def registry_access(cls, form: dict, session: str) -> str:
        if ParametersManager().offline_mode:
            logger.critical("You can't access the registry without Internet access.")
        try:
            data = await cls.__call_licenses_endpoint((await cls.__create_client()).functions,
                                                      cls.LicenseAction.RegistryAccess,
                                                      body=form,
                                                      auth_token=session)
            token = data.get("token")
            if token is None:
                logger.error("Exegol server returned an unexpected response. Please retry later.")
                raise CancelOperation
            return token
        except ConnectError as e:
            logger.critical("Exegol server is unreachable. Do you have internet access?")
            raise e
        except (FunctionsRelayError, TransportError) as e:
            logger.critical("Exegol server seems to be unavailable for now. Please retry later.")
            raise e
        except FunctionsHttpError as e:
            if e.status == 403:
                logger.critical("Your license does not allow you to download this image.")
            elif e.status == 500:
                logger.critical("The Exegol servers are currently unavailable, please try again later.")
            else:
                logger.debug(f"Error code {e.status}")
                logger.critical(f"An error occurred for registry access: {e.message}")
            raise e

    ## Public DB ##
    ### images table ###

    @classmethod
    async def list_all_images(cls, arch: str) -> List[SupabaseImage]:
        if ParametersManager().offline_mode:
            logger.warning("Can't list images without Internet access. Skipping.")
            return []
        logger.debug(f"Listing images from metadata table")
        image_list = await cls.__execute((await cls.__create_client())
                                         .table("images")
                                         .select("*")
                                         .eq("arch", arch)
                                         .not_.is_("version", None))
        supa_images: List[SupabaseImage] = []
        if image_list is not None:
            # Pydantic validation
            for x in image_list.data:
                supa_images.append(SupabaseImage.model_validate(x))

        return supa_images

    @classmethod
    async def get_tag_version(cls, tag: str, arch: str) -> Tuple[Optional[str], Optional[str]]:
        """Get the latest version information of a specific image"""
        if ParametersManager().offline_mode:
            return None, None
        logger.debug(f"Fetching latest version of a specific image {tag} from metadata table")
        image: Optional[APIResponse] = await cls.__execute((await cls.__create_client())
                                                           .table("images")
                                                           .select("repo_digest, version")
                                                           .eq("arch", arch)
                                                           .eq("tag", tag)
                                                           .not_.is_("version", None)
                                                           .limit(1)
                                                           .maybe_single())
        if image is None:
            return None, None
        result: Optional[Dict[str, str]] = cast(Optional[Dict[str, str]], image.data)  # single mode cast
        if result is None:
            return None, None
        return result["repo_digest"], result["version"]



import os
from datetime import datetime
from time import sleep
from typing import List, Optional, Union, cast, Tuple, Set

import docker
import requests.exceptions
from docker import DockerClient
from docker.errors import APIError, DockerException, NotFound, ImageNotFound
from docker.models.images import Image
from docker.models.networks import Network
from docker.models.volumes import Volume
from docker.types import IPAMPool, IPAMConfig
from requests import ReadTimeout

from exegol.config.ConstantConfig import ConstantConfig
from exegol.config.DataCache import DataCache
from exegol.config.EnvInfo import EnvInfo
from exegol.config.UserConfig import UserConfig
from exegol.console.ExegolStatus import ExegolStatus
from exegol.console.TUI import ExegolTUI
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.exceptions.ExegolExceptions import ObjectNotFound
from exegol.manager.TaskManager import TaskManager
from exegol.model.ExegolContainer import ExegolContainer
from exegol.model.ExegolContainerTemplate import ExegolContainerTemplate
from exegol.model.ExegolImage import ExegolImage
from exegol.model.ExegolNetwork import ExegolNetwork
from exegol.model.SupabaseModels import SupabaseImage
from exegol.utils.ExeLog import logger, ExeLog
from exegol.utils.MetaSingleton import MetaSingleton
from exegol.utils.NetworkUtils import NetworkUtils
from exegol.utils.SessionHandler import SessionHandler
from exegol.utils.SupabaseUtils import SupabaseUtils


# SDK Documentation : https://docker-py.readthedocs.io/en/stable/index.html


class DockerUtils(metaclass=MetaSingleton):

    def __init__(self) -> None:
        """Utility class between exegol and the Docker SDK"""
        try:
            # Connect Docker SDK to the local docker instance.
            # Docker connection setting is loaded from the user environment variables.
            self.__client: DockerClient = docker.from_env()
            # Check if the docker daemon is serving linux container
            self.__daemon_info = self.__client.info()
            if self.__daemon_info.get("OSType", "linux").lower() != "linux":
                logger.critical(
                    f"Docker daemon is not serving linux container ! Docker OS Type is: {self.__daemon_info.get('OSType', 'linux')}")
            EnvInfo.initData(self.__daemon_info)
        except DockerException as err:
            if 'ConnectionRefusedError' in str(err):
                logger.critical(f"Unable to connect to docker (from env config). Is docker running on your machine? Exiting.{os.linesep}"
                                f"    Check documentation for help: https://docs.exegol.com/troubleshooting#unable-to-connect-to-docker")
            elif 'FileNotFoundError' in str(err):
                logger.critical(f"Unable to connect to docker. Is docker installed on your machine? Exiting.{os.linesep}"
                                f"    Check documentation for help: https://docs.exegol.com/troubleshooting#unable-to-connect-to-docker")
            elif 'PermissionError' in str(err):
                logger.critical(f"Docker is installed on your host but you don't have the permission to interact with it. Exiting.{os.linesep}"
                                f"    Check documentation for help: https://docs.dev.exegol.com/first-install#_2-wrapper-install")
            else:
                logger.error(err)
                logger.critical(
                    "Unable to connect to docker (from env config). Is docker operational and accessible? on your machine? "
                    "Exiting.")
        except (ReadTimeout, requests.exceptions.ConnectionError):
            logger.critical("Docker daemon seems busy, Exegol receives timeout response. Try again later.")
        self.__images: Optional[List[ExegolImage]] = None
        self.__containers: Optional[List[ExegolContainer]] = None

    def clearCache(self) -> None:
        """Remove class's images and containers data cache
        Only needed if the list has to be updated in the same runtime at a later moment"""
        self.__containers = None
        self.__images = None

    def getDockerInfo(self) -> dict:
        """Fetch info from docker daemon"""
        return self.__daemon_info

    # # # Container Section # # #

    async def listContainers(self) -> List[ExegolContainer]:
        """List available docker containers.
        Return a list of ExegolContainer"""
        if self.__containers is None:
            logger.verbose("Loading Exegol containers")
            self.__containers = []
            try:
                docker_containers = self.__client.containers.list(all=True, filters={"name": "exegol-", "label": f"{ExegolImage.Labels.app.value}=Exegol"})
            except APIError as err:
                logger.debug(err)
                logger.critical(err.explanation)
                # Not reachable, critical logging will exit
                return  # type: ignore
            except ReadTimeout:
                logger.critical("Received a timeout error, Docker is busy... Unable to list containers, retry later.")
                return  # type: ignore
            for container in docker_containers:
                self.__containers.append(ExegolContainer(container))
        return self.__containers

    def createContainer(self, model: ExegolContainerTemplate, temporary: bool = False) -> ExegolContainer:
        """Create an Exegol container from an ExegolContainerTemplate configuration.
        Return an ExegolContainer if the creation was successful."""
        logger.info("Creating new exegol container")
        model.prepare()
        logger.debug(model)
        # Preload docker volume before container creation
        for volume in model.config.getVolumes():
            if volume.get('Type', '?') == "volume":
                docker_volume = self.__loadDockerVolume(volume_path=volume['Source'], volume_name=volume['Target'])
                if docker_volume is None:
                    logger.warning(f"Error while creating docker volume '{volume['Target']}'")
        entrypoint, command = model.config.getEntrypointCommand()
        logger.debug(f"Entrypoint: {entrypoint}")
        logger.debug(f"Cmd: {command}")
        # The 'create' function must be called to create a container without starting it
        # in order to hot patch the entrypoint.sh with wrapper features (the container will be started after postCreateSetup)
        docker_create_function = self.__client.containers.create
        docker_args = {"image": model.image.getDockerRef(),
                       "entrypoint": entrypoint,
                       "command": command,
                       "detach": True,
                       "name": model.getContainerName(),
                       "hostname": model.config.hostname,
                       "extra_hosts": model.config.getExtraHost(),
                       "devices": model.config.getDevices(),
                       "environment": model.config.getEnvs(),
                       "labels": model.config.getLabels(),
                       "ports": model.config.getPorts(),
                       "privileged": model.config.getPrivileged(),
                       "cap_add": model.config.getCapabilities(),
                       "sysctls": model.config.getSysctls(),
                       "shm_size": model.config.shm_size,
                       "stdin_open": model.config.interactive,
                       "tty": model.config.tty,
                       "mounts": model.config.getVolumes(),
                       "userns_mode": "host",
                       "working_dir": model.config.getWorkingDir()}
        # Add networking args
        if model.config.isNetworkDisabled():
            docker_args["network_disabled"] = True
        else:
            docker_args["network"], network_driver = model.config.getNetwork()
            if (docker_args["network"] is not None and
                    network_driver is not None and
                    not self.networkExist(cast(str, docker_args["network"]))):
                if not self.createNetwork(network_name=cast(str, docker_args["network"]), driver=network_driver):
                    logger.critical("Unable to create the dedicated network for the new container. Aborting.")

        # Handle temporary arguments
        if temporary:
            # Only the 'run' function support the "remove" parameter
            docker_create_function = self.__client.containers.run
            docker_args["remove"] = temporary
            docker_args["auto_remove"] = temporary

        # Create container
        try:
            container = docker_create_function(**docker_args)
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to create container, retry later.")
            raise RuntimeError
        except APIError as err:
            if err.explanation is None:
                err.explanation = ''
            elif type(err.explanation) is bytes:
                err.explanation = cast(bytes, err.explanation).decode('utf-8')
            message = err.explanation.replace('[', '\\[')
            if message is not None:
                message = message.replace('[', '\\[')
                logger.error(f"Docker error received: {message}")
            logger.debug(err)
            model.rollback()
            try:
                container = self.__client.containers.list(all=True, filters={"name": model.getContainerName(), "label": f"{ExegolImage.Labels.app.value}=Exegol"})
                if container is not None and len(container) > 0:
                    for c in container:
                        if c.name == model.getContainerName():  # Search for exact match
                            container[0].remove()
                            logger.debug("Container removed")
            except APIError as e:
                logger.debug(f"Error while removing dcontainer: {e}")
            try:
                if docker_args.get("network") is not None and self.removeNetwork(cast(str, docker_args["network"])):
                    logger.debug("Network removed")
            except (ValueError, APIError) as e:
                logger.debug(f"Error while removing dedicated network: {e}")
            logger.critical("Error while creating exegol container. Exiting.")
            raise RuntimeError
        if container is not None:
            logger.success("Exegol container successfully created!")
        else:
            logger.critical("Unknown error while creating exegol container. Exiting.")
            raise RuntimeError
        return ExegolContainer(container, model)

    def getContainer(self, tag: str) -> ExegolContainer:
        """Get an ExegolContainer from tag name."""
        try:
            # Fetch potential container match from DockerSDK
            container = self.__client.containers.list(all=True, filters={"name": f"exegol-{tag}", "label": f"{ExegolImage.Labels.app.value}=Exegol"})
        except APIError as err:
            logger.debug(err)
            logger.critical(err.explanation)
            # Not reachable, critical logging will exit
            return  # type: ignore
        # Check if there is at least 1 result. If no container was found, raise ObjectNotFound.
        if container is None or len(container) == 0:
            # Handle case-insensitive OS
            if EnvInfo.isWindowsHost() or EnvInfo.isMacHost():
                # First try to fetch the container as-is (for retroactive support with old container with uppercase characters)
                # If the user's input didn't match any container, try to force the name in lowercase if not already tried
                lowered_tag = tag.lower()
                if lowered_tag != tag:
                    return self.getContainer(lowered_tag)
            raise ObjectNotFound
        # Filter results with exact name matching
        for c in container:
            if c.name == f"exegol-{tag}":
                # When the right container is found, select it and stop the search
                return ExegolContainer(c)
        # When there is some close container's name,
        # docker may return some results but none of them correspond to the request.
        # In this case, ObjectNotFound is raised
        raise ObjectNotFound

    def removeContainerById(self, container_id: str) -> bool:
        """Get an ExegolContainer from its ID"""
        try:
            c = self.__client.containers.get(container_id)
        except NotFound:
            raise ObjectNotFound
        except APIError as err:
            logger.debug(err)
            logger.critical(err.explanation)
            raise RuntimeError
        c.remove()
        return True

    def isContainerExist(self, container_id: str) -> Optional[str]:
        """
        Test if a container still exists from its ID.
        :param container_id:
        :return: Return the name of the container if it still exists, None if it doesn't exist anymore.
        """
        try:
            c = self.__client.containers.get(container_id)
            return c.name
        except NotFound:
            return None
        except APIError as err:
            logger.debug(err)
            logger.critical(err.explanation)
        return None

    # # # Volumes Section # # #

    def __loadDockerVolume(self, volume_path: str, volume_name: str) -> Volume:
        """Load or create a docker volume for exegol containers
        (must be created before the container, SDK limitation)
        Return the docker volume object"""
        try:
            os.makedirs(volume_path, exist_ok=True)
        except PermissionError:
            logger.error("Unable to create the volume folder on the filesystem locally.")
            logger.critical(f"Insufficient permission to create the folder: {volume_path}")
        try:
            # Check if volume already exist
            volume = self.__client.volumes.get(volume_name)
            path = volume.attrs.get('Options', {}).get('device', '')
            if path != volume_path:
                try:
                    self.__client.api.remove_volume(name=volume_name)
                    raise NotFound('Volume must be reloaded')
                except APIError as e:
                    if e.status_code == 409:
                        logger.warning("The path of the volume specified by the user is not the same as in the existing docker volume. "
                                       "The user path will be [red]ignored[/red] as long as the docker volume already exists.")
                        logger.verbose("The volume is already used by some container and cannot be automatically removed.")
                        logger.debug(e.explanation)
                    else:
                        raise NotFound('Volume must be reloaded')
                except ReadTimeout:
                    logger.error(f"Received a timeout error, Docker is busy... Volume {volume_name} cannot be automatically removed. Please, retry later the following command:{os.linesep}"
                                 f"    [orange3]docker volume rm {volume_name}[/orange3]")
        except NotFound:
            try:
                # Creating a docker volume bind to a host path
                # Docker volume are more easily shared by container
                # Docker volume can load data from container image on host's folder creation
                volume = self.__client.volumes.create(volume_name, driver="local",
                                                      driver_opts={'o': 'bind',
                                                                   'device': volume_path,
                                                                   'type': 'none'})
            except APIError as err:
                logger.error(f"Error while creating docker volume '{volume_name}'.")
                logger.debug(err)
                logger.critical(err.explanation)
                raise RuntimeError
            except ReadTimeout:
                logger.critical(f"Received a timeout error, Docker is busy... Volume {volume_name} cannot be created.")
                raise RuntimeError
        except APIError as err:
            logger.critical(f"Unexpected error by Docker SDK : {err}")
            raise RuntimeError
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to enumerate volume, retry later.")
            raise RuntimeError
        return volume

    # # # Network Section # # #

    def listAttachableNetworks(self) -> List[Network]:
        """List every non-default networks"""
        networks = []
        try:
            networks = self.__client.networks.list()
        except APIError as e:
            raise e
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to enumerate volume, retry later.")
        for net in networks.copy():
            if net.name in ExegolNetwork.DEFAULT_DOCKER_NETWORK:
                networks.remove(net)
        return networks

    def listExegolNetworks(self) -> List[Network]:
        """List every exegol networks"""
        networks = []
        try:
            networks = self.__client.networks.list(filters={"label": "source=exegol"})
        except APIError as e:
            raise e
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to enumerate volume, retry later.")
        return networks

    def getNetwork(self, network_name: str, exegol_only: bool = False) -> Optional[Network]:
        """Find a specific network"""
        networks: List[Network] = []
        filter = {}
        if exegol_only:
            filter["label"] = "source=exegol"
        try:
            networks = self.__client.networks.list(names=network_name, filters=filter)
        except APIError as e:
            raise e
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to enumerate volume, retry later.")
        for net in networks:
            # Search for an exact match
            if net.name == network_name:
                return net
        return None

    def networkExist(self, network_name: str) -> bool:
        """Return True is the supplied network name exist"""
        return self.getNetwork(network_name) is not None

    def __listDockerNetworks(self) -> List[str]:
        """List every docker network"""
        networks = []
        try:
            networks = self.__client.networks.list()
        except APIError as e:
            raise e
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to enumerate volume, retry later.")
        networks_range = []
        for net in networks:
            net_config = net.attrs.get('IPAM', {}).get('Config', [])
            if net_config is not None and len(net_config) > 0:
                net_range = net_config[0].get('Subnet')
                if net_range:
                    networks_range.append(net_range)

        return networks_range

    def createNetwork(self, network_name: str, driver: str) -> bool:
        """Create a new exegol network"""
        docker_networks = self.__listDockerNetworks()
        ip_pool = IPAMPool(subnet=str(NetworkUtils.get_next_available_range(
            UserConfig().network_dedicated_range,
            UserConfig().network_default_netmask,
            docker_networks)))
        config = IPAMConfig(pool_configs=[ip_pool])
        try:
            network: Network = self.__client.networks.create(name=network_name, driver=driver, labels={"source": "exegol"}, check_duplicate=True, ipam=config)
            return True
        except APIError as e:
            if e.status_code == 409:
                logger.error("This network already exist.")
                return False
            raise e
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to enumerate volume, retry later.")
        return False

    def removeNetwork(self, network_name: Optional[str] = None, network: Optional[Network] = None) -> bool:
        """Remove exegol network"""
        if network is None:
            if network_name is None:
                raise ValueError("One of the parameter must be supplied.")
            elif network_name in ExegolNetwork.DEFAULT_DOCKER_NETWORK:
                # Default docker driver cannot be deleted
                return False
            network = self.getNetwork(network_name, exegol_only=True)

        if network is not None:
            if network.name in ExegolNetwork.DEFAULT_DOCKER_NETWORK:
                # Default docker driver cannot be deleted
                return False
            try:
                network.remove()
                logger.success(f"Dedicated network successfully removed.")
                return True
            except NotFound:
                logger.verbose(f"The dedicated network {network.name} was already removed.")
            except APIError as e:
                logger.error(f"The associated dedicated network cannot be automatically removed. "
                             f"You have to delete it manually ({network.name}). Error: {e.explanation}")
        else:
            logger.info(f"The network {network_name} will not be deleted. Only exegol network can be automatically deleted.")
        return False

    # # # Image Section # # #

    async def listImages(self, include_version_tag: bool = False, include_locked: bool = False, include_custom: bool = False) -> List[ExegolImage]:
        """List available docker images.
        Return a list of ExegolImage"""
        if self.__images is None:
            logger.verbose("Loading Exegol images")
            async with ExegolStatus(f"Loading Exegol images", spinner_style="blue") as s:
                TaskManager.add_task(
                    SupabaseUtils.list_all_images(ParametersManager().arch),
                    TaskManager.TaskId.RemoteImageList)
                TaskManager.add_task(
                    self.__listOfficialLocalImages(),
                    TaskManager.TaskId.LocalImageList)
                remote_images: List[SupabaseImage]
                local_images: List[Image]
                remote_images, local_images = await TaskManager.gather(TaskManager.TaskId.RemoteImageList, TaskManager.TaskId.LocalImageList)
                if include_custom and len(UserConfig().custom_images) > 0:
                    s.update(status=f"Retrieving [green]custom[/green] images")
                    for custom in UserConfig().custom_images:
                        local_images.extend(await self.__listCustomLocalImages(custom))
                    logger.verbose("Retrieved [green]custom[/green] images")
                self.__images = ExegolImage.mergeImages(remote_images, local_images)
        result = self.__images
        assert result is not None
        # Caching latest images
        TaskManager.add_task(
            DataCache().update_image_cache([img for img in result if not img.isVersionSpecific()])
        )
        if not (logger.isEnabledFor(ExeLog.VERBOSE) or include_locked):
            # ToBeRemoved images are only shown in verbose mode
            result = [i for i in result if not i.isLocked()]
        if not include_version_tag:
            # Version specific images not installed are excluded by default
            result = [img for img in result if not img.isVersionSpecific() or img.isInstall()]
        return result

    async def getOfficialImageFromList(self, tag: str) -> Union[ExegolImage, str]:
        """Get an ExegolImage from tag name."""
        # Fetch every official images available
        images = await self.listImages(include_version_tag=True, include_locked=True)
        match: Optional[ExegolImage] = None
        # image tag without version
        tag_only = tag.split('-')[0] if '-' in tag else None
        tag_only_match = None
        # Find a match
        for i in images:
            if i.getName() == tag:
                # If there is a locked image keep it as default
                if i.isLocked():
                    match = i
                else:
                    # Return the first non-outdated image
                    return i
            elif tag_only is not None and i.getName() == tag_only:
                tag_only_match = i
        # If there is any match without lock (outdated) status, return the last outdated image found.
        if match is not None:
            return match
        # If the version specific image is not found, return the latest version
        if tag_only_match is not None:
            tag_only_match.setupAsLegacy(tag)
            return tag_only_match
        # If there is no match at all, raise ObjectNotFound to handle the error
        raise ObjectNotFound

    def __getImage(self, image_name: str) -> Optional[Image]:
        """Get a specific local image from docker"""
        logger.debug(f"Trying to get {image_name} from docker")
        try:
            return self.__client.images.get(image_name)
            # DockerSDK image get is an exact matching, no need to add more check
        except APIError as err:
            if err.status_code == 404:
                pass
            else:
                logger.critical(f"Error on image loading: {err}")
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to list images, retry later.")
        return None

    async def getInstalledImage(self, tag: str, repository: Optional[str] = None, skip_cache: bool = False) -> ExegolImage:
        """Get an already installed ExegolImage from tag name."""
        try:
            if not skip_cache and self.__images is not None:
                # Try to find image from cache
                for img in self.__images:
                    if img.getName() == tag:
                        if not img.isInstall() or not img.isUpToDate():
                            # Refresh local image status in case of installation/upgrade operations
                            self.__findImageMatch(img)
                        return img
            # Load image
            docker_local_image = None
            search = [f"{i}:{tag}" for i in ConstantConfig.OFFICIAL_REPOSITORIES] if repository is None else [f"{repository}:{tag}"]
            if "/" in tag and SessionHandler().enterprise_feature_access():
                search.insert(0, tag)
            for image_name in search:
                docker_local_image = self.__getImage(image_name)
                if docker_local_image is not None:
                    break
            if docker_local_image is None:
                # try to find it in recovery mode
                logger.verbose("Unable to find your image. Trying to find in recovery mode.")
                recovery_images = self.__findLocalRecoveryImages(include_untag=True)
                match = []
                for img in recovery_images:
                    if ExegolImage.parseAliasTagName(img) == tag:
                        match.append(ExegolImage(docker_image=img))
                if len(match) == 1:
                    return match[0]
                elif len(match) > 1:
                    return cast(ExegolImage, await ExegolTUI.selectFromTable(match, object_type=ExegolImage))
                raise ObjectNotFound
            return await ExegolImage(docker_image=docker_local_image).autoLoad()
        except ObjectNotFound:
            logger.critical(f"The desired image is not installed or do not exist ({repository + ':' if repository else ''}{tag}). Exiting.")
        return  # type: ignore

    async def __listLocalImages(self, image_name: str, tag: Optional[str] = None) -> Tuple[List[Image], Set[str]]:
        logger.debug("Fetching local image tags, digests (and other attributes)")
        try:
            image_full_name = image_name + ("" if tag is None else f":{tag}")
            images = self.__client.images.list(image_full_name, filters={"dangling": False, "label": [f"{ExegolImage.Labels.app.value}=Exegol"]})
        except APIError as err:
            logger.debug(err)
            logger.critical(err.explanation)
            # Not reachable, critical logging will exit
            return  # type: ignore
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to list local images, retry later.")
            return  # type: ignore
        # Filter out image non-related to the right repository
        result = []
        ids = set()
        for img in images:
            # len tags = 0 handle exegol <none> images (nightly image lost their tag after update)
            if len(img.attrs.get('RepoTags', [])) == 0 or \
                    image_name in [repo_tag.split(':')[0] for repo_tag in img.attrs.get("RepoTags", [])]:
                result.append(img)
                ids.add(img.id)
        return result, ids

    async def __listOfficialLocalImages(self, tag: Optional[str] = None) -> List[Image]:
        """List local docker images already installed.
        Return a list of docker images objects"""
        result = []
        ids = set()
        for repository in ConstantConfig.OFFICIAL_REPOSITORIES:
            sub_result, sub_ids = await self.__listLocalImages(repository, tag)
            result.extend(sub_result)
            ids.update(sub_ids)

        # Try to find lost Exegol images
        recovery_images = self.__findLocalRecoveryImages()
        for img in recovery_images:
            # Docker can keep track of 2 images maximum with RepoTag or RepoDigests, after it's hard to track origin without labels, so this recovery option is "best effort"
            if img.id in ids:
                # Skip image from other repo and image already found
                logger.debug(f"Duplicate found in recovery mode! {img}")
                continue
            else:
                result.append(img)
                ids.add(img.id)
        return result

    async def __listCustomLocalImages(self, image_name: str, tag: Optional[str] = None) -> List[Image]:
        """List local docker images already installed.
        Return a list of docker images objects"""
        result, _ = await self.__listLocalImages(image_name, tag)
        return result

    def __findLocalRecoveryImages(self, include_untag: bool = False) -> List[Image]:
        """This method try to recovery untagged docker images.
        Set include_untag option to recover images with a valid RepoDigest (no not dangling) but without tag."""
        try:
            # Try to find lost Exegol images
            recovery_images = self.__client.images.list(filters={"dangling": True})
            if include_untag:
                for image_name in ConstantConfig.OFFICIAL_REPOSITORIES:
                    recovery_images += self.__client.images.list(image_name, filters={"dangling": False})
        except APIError as err:
            logger.debug(f"Error occurred in recovery mode: {err}")
            return []
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to enumerate lost images, retry later.")
            return  # type: ignore
        result = []
        id_list = set()
        for img in recovery_images:
            # Docker can keep track of 2 images maximum with RepoTag or RepoDigests, after it's hard to track origin without labels, so this recovery option is "best effort"
            repo_tags = img.attrs.get('RepoTags')
            repo_digest = img.attrs.get('RepoDigests')
            if repo_tags is not None and len(repo_tags) > 0 or (not include_untag and repo_digest is not None and len(repo_digest) > 0) or img.id in id_list:
                # Skip image from other repo and image already found
                continue
            if img.labels is not None and img.labels.get(ExegolImage.Labels.app.value, '') == "Exegol":
                result.append(img)
                id_list.add(img.id)
        return result

    def __findImageMatch(self, remote_image: ExegolImage) -> None:
        """From a Remote ExegolImage, try to find a local match (using Remote DigestID).
        This method is useful if the image repository name is also lost"""
        remote_id = remote_image.getLatestRemoteId()
        if not remote_id:
            logger.debug("Latest remote id is not available... Falling back to the current remote id.")
            remote_id = remote_image.getRemoteId()
        try:
            logger.debug(f"Search image match for {remote_image.getRepository()}@{remote_id}")
            docker_image = self.__client.images.get(f"{remote_image.getRepository()}@{remote_id}")
        except ImageNotFound:
            # Fallback to get from tag for legacy image
            docker_image = self.__getImage(remote_image.getFullName())
            if docker_image is None:
                raise ObjectNotFound
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to find a specific image, retry later.")
            raise RuntimeError
        remote_image.resetDockerImage()
        remote_image.setDockerObject(docker_image)

    async def downloadImage(self, image: ExegolImage, install_mode: bool = False) -> bool:
        """Download/pull an ExegolImage"""
        if ParametersManager().offline_mode:
            logger.critical("It's not possible to download a docker image in offline mode ...")
            return False
        if "Unknown" in image.getStatus():
            logger.warning("This image cannot be downloaded at this time. Please try again later.")
            return False
        auth_config: Optional[dict] = None
        if await image.pullAuthNeeded():
            auth_config = await SessionHandler().get_registry_auth(image.getRepository(), image.getName())
        # Switch to install mode if the selected image is not already installed
        install_mode = install_mode or not image.isInstall()
        logger.info(f"{'Installing' if install_mode else 'Updating'} exegol image : {image.getName()}")
        name = image.updateCheck()
        if name is not None:
            download_size = '' if 'N/A' in image.getDownloadSize() else f" a [cyan1]~{image.getDownloadSize()}[/cyan1]"
            logger.info(f"Pulling compressed image, starting{download_size} download :satellite:")
            logger.info(f"Once downloaded and uncompressed, the image will take [cyan1]~{image.getRealSizeRaw()}[/cyan1] on disk :floppy_disk:")
            logger.debug(f"Downloading {image.getRepository()}:{name} ({image.getArch()})")
            try:
                await ExegolTUI.downloadDockerLayer(
                    self.__client.api.pull(repository=image.getRepository(),
                                           tag=name,
                                           stream=True,
                                           decode=True,
                                           platform="linux/" + image.getArch(),
                                           auth_config=auth_config))
                logger.success(f"Image successfully {'installed' if install_mode else 'updated'}")
                # Remove old image
                if not install_mode and image.isInstall() and UserConfig().auto_remove_images:
                    await self.removeImage(image, upgrade_mode=not install_mode)
                return True
            except APIError as err:
                if err.status_code == 500:
                    if err.explanation == "unauthorized":
                        logger.error("Permission denied! You cannot download this image, you need a valid Exegol license.")
                    else:
                        logger.error(f"Error: {err.explanation}")
                        logger.error(f"Error while contacting docker registry. Aborting.")
                elif err.status_code == 404:
                    if image.getRepository() == ConstantConfig.IMAGE_NAME and image.getName() != 'nightly':
                        logger.warning("This image version was not found, retrying on legacy registry.")
                        image.setRepository(ConstantConfig.COMMUNITY_IMAGE_NAME)
                        return await self.downloadImage(image, install_mode=install_mode)
                    logger.critical(f"The image has not been found on the docker registry: {err.explanation}")
                else:
                    logger.debug(f"Error: {err}")
                    logger.critical(f"An error occurred while downloading this image: {err.explanation}")
            except ReadTimeout:
                logger.critical(f"Received a timeout error, Docker is busy... Unable to download {name} image, retry later.")
        return False

    async def downloadVersionTag(self, image: ExegolImage) -> Union[ExegolImage, str]:
        """Pull a docker image for a specific version tag and return the corresponding ExegolImage"""
        if ParametersManager().offline_mode:
            logger.critical("It's not possible to download a docker image in offline mode ...")
            return ""
        auth_config: Optional[dict] = None
        if await image.pullAuthNeeded():
            auth_config = await SessionHandler().get_registry_auth(image.getRepository(), image.getLatestVersionName())
        try:
            image = self.__client.images.pull(repository=image.getRepository(),
                                              tag=image.getLatestVersionName(),
                                              platform="linux/" + image.getArch(),
                                              auth_config=auth_config)
            return ExegolImage(docker_image=image, isUpToDate=True)
        except APIError as err:
            if err.status_code == 500:
                return f"error while contacting docker registry: {err.explanation}"
            elif err.status_code == 404:
                return f"matching tag doesn't exist: {err.explanation}"
            else:
                logger.debug(f"Error: {err}")
                return f"en unknown error occurred while downloading this image : {err.explanation}"
        except ReadTimeout:
            logger.critical(f"Received a timeout error, Docker is busy... Unable to download an image tag, retry later the following command:{os.linesep}"
                            f"    [orange3]docker pull --platform linux/{image.getArch()} {image.getRepository()}:{image.getLatestVersionName()}[/orange3].")
            return  # type: ignore

    async def removeImage(self, image: ExegolImage, upgrade_mode: bool = False, silent_error: bool = False) -> bool:
        """Remove an ExegolImage from disk"""
        tag = image.removeCheck()
        if tag is None:  # Skip removal if image is not installed locally.
            return False
        async with ExegolStatus(f"Removing {'previous ' if upgrade_mode else ''}image [green]{image.getName()}[/green]...", spinner_style="blue"):
            try:
                if not image.isVersionSpecific() and image.getInstalledVersionName() != image.getName() and not upgrade_mode:
                    # Docker can't remove multiple images at the same tag, version specific tag must be remove first
                    logger.debug(f"Removing image {image.getFullVersionName()}")
                    if not self.__remove_image(image.getFullVersionName()):
                        logger.critical(f"An error occurred while removing this image : {image.getFullVersionName()}")
                logger.debug(f"Removing image {image.getLocalId()} ({image.getFullVersionName() if upgrade_mode else image.getFullName()})")
                if self.__remove_image(image.getLocalId()):
                    logger.verbose(f"Removing {'previous ' if upgrade_mode else ''}image [green]{image.getName()}[/green]...")
                    logger.success(f"{'Previous d' if upgrade_mode else 'D'}ocker image successfully removed.")
                    return True
            except APIError as err:
                # Handle docker API error code
                logger.verbose(err.explanation)
                if silent_error and not logger.isEnabledFor(ExeLog.VERBOSE):
                    return False
                if err.status_code == 409:
                    if upgrade_mode:
                        logger.error(f"The '{image.getName()}' image cannot be deleted yet, "
                                     "all containers using this old image must be deleted first.")
                    else:
                        logger.error(f"The '{image.getName()}' image cannot be deleted because "
                                     f"it is currently used by a container. Aborting.")
                elif err.status_code == 404:
                    logger.error(f"This image doesn't exist locally {image.getLocalId()} ({image.getFullName()}). Aborting.")
                else:
                    logger.critical(f"An error occurred while removing this image : {err}")
        return False

    def __remove_image(self, image_name: str) -> bool:
        """
        Handle docker image removal with timeout support
        :param image_name: Name of the docker image to remove
        :return: True is removal successful and False otherwise
        """
        try:
            self.__client.images.remove(image_name, force=False, noprune=False)
            return True
        except (ReadTimeout, requests.exceptions.ConnectionError):
            logger.warning("The deletion of the image has timeout. Docker is still processing the removal, please wait.")
            max_retry = 10
            wait_time = 5
            for i in range(5):
                try:
                    _ = self.__client.images.get(image_name)
                    # DockerSDK image getter is an exact matching, no need to add more check
                except APIError as err:
                    if err.status_code == 404:
                        return True
                    else:
                        logger.debug(f"Unexpected error after timeout: {err}")
                except (ReadTimeout, requests.exceptions.ConnectionError):
                    wait_time = wait_time + wait_time * i
                    logger.info(f"Docker timeout again ({i + 1}/{max_retry}). Next retry in {wait_time} seconds...")
                    sleep(wait_time)  # Wait x seconds before retry
            logger.error(f"The deletion of the image '{image_name}' has timeout, the deletion may be incomplete.")
        return False

    async def buildImage(self, tag: str, build_profile: Optional[str], build_dockerfile: Optional[str], dockerfile_path: str) -> None:
        """Build a docker image from source"""
        if ParametersManager().offline_mode:
            logger.critical("It's not possible to build a docker image in offline mode. The build process need access to internet ...")
            raise RuntimeError
        logger.info(f"Building exegol image : {tag}")
        if build_profile is None or build_dockerfile is None:
            build_profile = "full"
            build_dockerfile = "Dockerfile"
        logger.info("Starting build. Please wait, this will be long.")
        logger.verbose(f"Creating build context from [gold]{dockerfile_path}[/gold] with "
                       f"[green][b]{build_profile}[/b][/green] profile ({ParametersManager().arch}).")
        if EnvInfo.arch != ParametersManager().arch:
            logger.warning("Building an image for a different host architecture can cause unexpected problems and slowdowns!")
        try:
            # path is the directory full path where Dockerfile is.
            # tag is the name of the final build
            # dockerfile is the Dockerfile filename
            await ExegolTUI.buildDockerImage(
                self.__client.api.build(path=dockerfile_path,
                                        dockerfile=build_dockerfile,
                                        tag=f"{ConstantConfig.COMMUNITY_IMAGE_NAME}:{tag}",
                                        buildargs={"TAG": tag,
                                                   "VERSION": "local",
                                                   "BUILD_PROFILE": build_profile,
                                                   "BUILD_DATE": datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')},
                                        platform="linux/" + ParametersManager().arch,
                                        rm=True,
                                        forcerm=True,
                                        pull=True,
                                        decode=True))
            logger.success(f"Exegol image successfully built")
        except APIError as err:
            logger.debug(f"Error: {err}")
            if err.status_code == 500:
                logger.error(f"Error: {err.explanation}")
                logger.error("Error while contacting docker hub. You probably don't have internet. Aborting.")
                logger.debug(f"Error: {err}")
            else:
                logger.critical(f"An error occurred while building this image : {err}")
        except ReadTimeout:
            logger.critical("Received a timeout error, Docker is busy... Unable to build the local image, retry later.")
            raise RuntimeError



import os
from typing import Optional

from rich.console import JustifyMethod
from rich.highlighter import Highlighter
from rich.progress import TextColumn, Task, DownloadColumn
from rich.style import StyleType
from rich.table import Column
from rich.text import Text

from exegol.utils.ExeLog import logger


class LayerTextColumn(TextColumn, DownloadColumn):
    """Merging two Rich class to obtain a double behavior in the same RichTable"""

    def __init__(self,
                 text_format: str,
                 layer_key: str,
                 style: StyleType = "none",
                 justify: JustifyMethod = "left",
                 markup: bool = True,
                 highlighter: Optional[Highlighter] = None,
                 table_column: Optional[Column] = None,
                 binary_units: bool = False
                 ) -> None:
        # Custom field
        self.__data_key = layer_key
        # Inheritance configuration
        try:
            TextColumn.__init__(self, text_format, style, justify, markup, highlighter, table_column)
        except TypeError:
            logger.critical(f"Your version of Rich does not correspond to the project requirements. Please update your dependencies with pip:{os.linesep}"
                            f"[bright_magenta]python3 -m pip install --user --requirement requirements.txt[/bright_magenta]")
        DownloadColumn.__init__(self, binary_units, table_column)

    def render(self, task: "Task") -> Text:
        """Custom render depending on the existence of data with data_key"""
        if task.fields.get(self.__data_key) is None:
            # Default render with classic Text render
            return TextColumn.render(self, task)
        else:
            # If the task download a file, render the Download progress view
            return DownloadColumn.render(self, task)



import os
from argparse import Namespace
from typing import List, Optional, Tuple, Union, Dict, cast

from exegol.console.ConsoleFormat import richLen
from exegol.config.EnvInfo import EnvInfo
from exegol.utils.ExeLog import logger


class Option:
    """This object allows to define and configure an argparse parameter"""

    def __init__(self, *args, dest: Optional[str] = None, **kwargs) -> None:
        """Generic class to handle Key:Value object directly from the constructor"""
        # Set arguments to the object to save every setting, these values will be sent to the argparser
        self.args = args
        self.kwargs = kwargs
        if dest is not None:
            self.kwargs["dest"] = dest
        self.dest = dest

    def __repr__(self) -> str:
        """This overload allows to format the name of the object.
        Mainly used by developers to easily identify objects"""
        return f"Option: {str(self.dest) if self.dest is not None else self.kwargs.get('metavar', 'Option not found')}"


class GroupArg:
    """This object allows you to group a set of options within the same group"""

    def __init__(self, *options, title: Optional[str] = None, description: Optional[str] = None,
                 is_global: bool = False) -> None:
        self.title = title
        self.description = description
        self.options: Tuple[Dict[str, Union[Option, bool]]] = cast(Tuple[Dict[str, Union[Option, bool]]], options)
        self.is_global = is_global

    def __repr__(self) -> str:
        """This overload allows to format the name of the object.
        Mainly used by developers to easily identify objects"""
        return f"GroupArg: {self.title}"


class Command:
    """The Command class is the root of all CLI actions"""

    def __init__(self) -> None:
        # Root command usages (can be overwritten by subclasses to display different use cases)
        self._pre_usages = "[underline]To see specific examples run:[/underline][italic] exegol [cyan]command[/cyan] -h[/italic]"
        self._usages = {
            "Install an exegol image": "exegol install",
            "Open an exegol shell": "exegol start",
            "Show exegol images & containers": "exegol info",
            "Update an image": "exegol update",
            "See commands examples to execute": "exegol exec -h",
            "Remove a container": "exegol remove",
            "Uninstall an image": "exegol uninstall",
            "Build a local exegol image": "exegol build",
            "Stop a container": "exegol stop"
        }
        self._post_usages = ""

        # Name of the object
        self.name = type(self).__name__.lower()
        # Global parameters
        # Only used by dockerhub registry
        #self.verify = Option("-k", "--insecure",
        #                     dest="verify",
        #                     action="store_false",
        #                     default=True,
        #                     required=False,
        #                     help="Allow insecure server connections for web requests, "
        #                          "e.g. when fetching info from DockerHub "
        #                          "(default: [green]Secure[/green])")
        self.quiet = Option("-q", "--quiet",
                            dest="quiet",
                            action="store_true",
                            default=False,
                            help="Show no information at all")
        self.verbosity = Option("-v", "--verbose",
                                dest="verbosity",
                                action="count",
                                default=0,
                                help="Verbosity level (-v for verbose, -vv for advanced, -vvv for debug)")
        self.arch = Option("--arch",
                           dest="arch",
                           action="store",
                           choices={"amd64", "arm64"},
                           default=EnvInfo.arch,
                           help=f"Overwrite default image architecture (default, host's arch: [blue]{EnvInfo.arch}[/blue])")
        self.offline_mode = Option("--offline",
                                   dest="offline_mode",
                                   action="store_true",
                                   help=f"Run exegol in offline mode, no request will be made on internet (default: [red]Disable[/red])")
        self.accept_eula = Option("--accept-eula",
                                  action="store_true",
                                  dest="accept_eula",
                                  help="Accept the End-User License Agreement of Exegol")
        # TODO review non-interactive mode
        # self.interactive_mode = Option("--non-interactive",
        #                               dest="interactive_mode",
        #                               action="store_false",
        #                               help="[red](WIP)[/red] Prevents Exegol from interactively requesting information. "
        #                                    "If critical information is missing, an error will be raised.")

        # Main global group of argparse
        self.groupArgs = [
            GroupArg({"arg": self.verbosity, "required": False},
                     # {"arg": self.interactive_mode, "required": False},
                     {"arg": self.quiet, "required": False},
                     #{"arg": self.verify, "required": False},
                     {"arg": self.offline_mode, "required": False},
                     {"arg": self.accept_eula, "required": False},
                     {"arg": self.arch, "required": False},
                     title="[blue]Optional arguments[/blue]",
                     is_global=True)
        ]

    def __call__(self, *args, **kwargs):
        """This method is called by the main controller (ExegolController)
        to get the function, execute it and launch the action.
        This method must be overloaded in all child classes to ensure the correct execution of the thread"""
        logger.debug("The called command is : ", self.name)
        logger.debug("the object is", type(self).__name__)
        raise NotImplementedError

    def __repr__(self) -> str:
        """This overload allows to format the name of the object.
        Mainly used by developers to easily identify objects"""
        return self.name

    def populate(self, args: Namespace) -> None:
        """This method replaces the parsing objects (Option) with the result of the parsing"""
        for arg in vars(args).keys():
            # Check if the argument exist in the current class
            if arg in self.__dict__:
                # If so, overwrite it with the corresponding value after parsing
                self.__setattr__(arg, vars(args)[arg])

    def check_parameters(self) -> List[str]:
        """This method identifies the missing required parameters"""
        missingOption = []
        for groupArg in self.groupArgs:
            for option in groupArg.options:
                if option.get("required", False):
                    data = option.get("arg")
                    assert data is not None and type(data) is Option
                    if data.dest is not None and self.__dict__.get(data.dest) is None:
                        missingOption.append(data.dest)
        return missingOption

    def formatEpilog(self) -> str:
        epilog = "[blue]Examples:[/blue]" + os.linesep
        epilog += self._pre_usages + os.linesep
        keys_len = {}
        # Replace [.*] rich tag for line length count
        for k in self._usages.keys():
            keys_len[k] = richLen(k)
        max_key = max(keys_len.values())
        for k, v in self._usages.items():
            space = ' ' * (max_key - keys_len.get(k, 0) + 2)
            epilog += f"  {k}:{space}[i]{v}[/i]{os.linesep}"
        epilog += self._post_usages + os.linesep
        return epilog



import ipaddress
from typing import Union, Set, List, Optional

import ifaddr
from rich.prompt import Prompt

from exegol.model.ExegolNetwork import ExegolNetworkMode
from exegol.utils.ExeLog import logger, console


class NetworkUtils:
    adapters = ifaddr.get_adapters()

    @classmethod
    def get_host_addresses(cls) -> Set[str]:
        result = set()
        for adapter in cls.adapters:
            for ip in adapter.ips:
                # Filter for IPv4
                if ip.is_IPv4:
                    result.add(str(ip.ip))
        return result

    @classmethod
    def get_default_large_range(cls) -> ipaddress.IPv4Network:
        available_networks = []
        for net in ipaddress.IPv4Network("172.16.0.0/12").subnets(new_prefix=16):
            if not cls.__test_conflict(net):
                available_networks.append(net)
        if len(available_networks) > 0:
            # Choose last network available to avoid conflict with docker
            selected_network = available_networks[-1]
        else:
            logger.error("Unable to automatically find an available network !")
            selected_network = cls.__user_select_network()
        logger.debug(f"Using exegol global network: {selected_network}")
        return selected_network

    @classmethod
    def get_default_large_range_text(cls) -> str:
        return str(cls.get_default_large_range())

    @classmethod
    def __user_select_network(cls) -> ipaddress.IPv4Network:
        while True:
            net_input = Prompt.ask(prompt="Please choose manually a network range for dedicated exegol networks (CIDR format):", console=console)
            try:
                net_user = ipaddress.IPv4Network(net_input)
                if not net_user.is_private:
                    logger.error("Only private IP can be used.")
                else:
                    number_subnet = len([n for n in net_user.subnets(new_prefix=28)])
                    if cls.__test_conflict(net_user):
                        logger.error("This network is already in use on your machine. Please choose another available network.")
                    else:
                        if number_subnet <= 16:
                            logger.warning(f"The selected network can only have {number_subnet} exegol dedicated networks. You can change this setting from the Exegol config file.")
                        return net_user
            except ipaddress.AddressValueError:
                logger.error("The supplied network isn't a valid IPv4 network")
            except ipaddress.NetmaskValueError:
                logger.error("The supplied netmask isn't valid for an IPv4 address")
            except ValueError:
                logger.error("The supplied IPv4 isn't a network address.")

    @classmethod
    def __test_conflict(cls, network: ipaddress.IPv4Network, docker_networks: Optional[List[str]] = None) -> bool:
        """Test if the supplied network has conflict with any existing network card on the host"""
        for adapter in cls.adapters:
            # print("IPs of network adapter " + adapter.nice_name)
            for ip in adapter.ips:
                if ip.is_IPv6:
                    continue
                if ipaddress.ip_network(f"{ip.ip}/{ip.network_prefix}", strict=False).overlaps(network):
                    return True
        if docker_networks is not None:
            for docker_net in docker_networks:
                if ipaddress.ip_network(docker_net, strict=False).overlaps(network):
                    return True
        return False

    @classmethod
    def get_next_available_range(cls, large_range: str, network_size: int, docker_subnets: List[str]) -> ipaddress.IPv4Network :
        exegol_network = ipaddress.IPv4Network(large_range)
        for net in exegol_network.subnets(new_prefix=network_size):
            if not cls.__test_conflict(net, docker_subnets):
                return net
        raise ValueError("No more sub-network available. Please remove some network or change the exegol network configuration.")

    @classmethod
    def parse_netmask(cls, netmask: Union[int, str], default: int) -> int:
        if type(netmask) is int:
            if 32 >= netmask >= 0:
                return netmask
        else:
            assert type(netmask) is str
            if netmask.startswith("/"):
                netmask = netmask[1:]
            try:
                result = int(netmask)
                if 0 > result > 32:
                    raise ValueError()
                return result
            except ValueError:
                try:
                    return ipaddress.IPv4Network(f"0.0.0.0/{netmask}").prefixlen
                except ipaddress.NetmaskValueError:
                    pass
        logger.error(f"The supplied netmask is invalid (must be IPv4): {netmask}")
        return default

    @classmethod
    def get_options(cls) -> Set[str]:
        options = [x.name for x in ExegolNetworkMode if x != ExegolNetworkMode.attached]

        return set(options)



from pathlib import Path
from typing import List, cast

from exegol.config.ConstantConfig import ConstantConfig
from exegol.console.ConsoleFormat import boolFormatter
from exegol.model.ExegolNetwork import ExegolNetworkMode
from exegol.utils.DataFileUtils import DataFileUtils
from exegol.utils.ExeLog import logger
from exegol.utils.MetaSingleton import MetaSingleton
from exegol.utils.NetworkUtils import NetworkUtils
from exegol.utils.SessionHandler import SessionHandler


class UserConfig(DataFileUtils, metaclass=MetaSingleton):
    """This class allows loading user defined configurations"""

    # Static choices
    start_shell_options = {'zsh', 'bash', 'tmux'}
    shell_logging_method_options = {'script', 'asciinema'}
    desktop_available_proto = {'http', 'vnc'}

    def __init__(self) -> None:
        # Defaults User config
        self.private_volume_path: Path = ConstantConfig.exegol_config_path / "workspaces"
        self.my_resources_path: Path = ConstantConfig.exegol_config_path / "my-resources"
        self.exegol_resources_path: Path = self.__default_resource_location('exegol-resources')
        self.exegol_images_path: Path = self.__default_resource_location('exegol-images')
        # Config
        self.auto_check_updates: bool = True
        self.auto_remove_images: bool = True
        self.auto_update_workspace_fs: bool = False
        self.default_start_shell: str = "zsh"
        self.enable_exegol_resources: bool = True
        # Shell logging
        self.shell_logging_method: str = "asciinema"
        self.shell_logging_compress: bool = True
        self.always_enable_shell_logging: bool = False
        # Desktop
        self.desktop_default_enable: bool = False
        self.desktop_default_localhost: bool = True
        self.desktop_default_proto: str = "http"
        # Network
        self.network_default_mode: str = ExegolNetworkMode.host.name
        self.network_fallback_mode: str = "nat"
        self.network_dedicated_range: str = ""  # Finding a default network can require a user interaction, loading only if needed from NetworkUtils.get_default_large_range_text().
        self.network_default_netmask: int = 28
        # Custom
        self.custom_images: List[str] = []

        # Dynamic default config
        dynamic_default = {"exegol_dedicated_range": NetworkUtils.get_default_large_range_text}
        super().__init__("config.yml", "yml", dynamic_default)

    def _build_file_content(self) -> str:
        # Dynamic default (if not already defined)
        if not self.network_dedicated_range:
            self.network_dedicated_range = cast(str, self._get_dynamic_default("exegol_dedicated_range"))

        # Config builder
        config = f"""# Exegol configuration
# Full documentation: https://docs.exegol.com/wrapper/configuration#configuration-file

# Volume path can be changed at any time but existing containers will not be affected by the update
volumes:
    # The my-resources volume is a storage space dedicated to the user to customize his environment and tools. This volume can be shared across all exegol containers.
    # Attention! The permissions of this folder (and subfolders) will be updated to share read/write rights between the host (user) and the container (root). Do not modify this path to a folder on which the permissions (chmod) should not be modified.
    my_resources_path: {self.my_resources_path}
    
    # Exegol resources are data and static tools downloaded in addition to docker images. These tools are complementary and are accessible directly from the host.
    exegol_resources_path: {self.exegol_resources_path}
    
    # Exegol images are the source of the exegol environments. These sources are needed when locally building an exegol image.
    exegol_images_path: {self.exegol_images_path}
    
    # When containers do not have an explicitly declared workspace, a dedicated folder will be created at this location to share the workspace with the host but also to save the data after deleting the container
    private_workspace_path: {self.private_volume_path}

config:
    # Enables automatic check for wrapper updates
    auto_check_update: {self.auto_check_updates}
    
    # Automatically remove outdated image when they are no longer used
    auto_remove_image: {self.auto_remove_images}
    
    # Automatically modifies the permissions of folders and sub-folders in your workspace by default to enable file sharing between the container with your host user.
    auto_update_workspace_fs: {self.auto_update_workspace_fs}
    
    # Default shell command to start
    default_start_shell: {self.default_start_shell}
    
    # Enable Exegol resources
    enable_exegol_resources: {self.enable_exegol_resources}
    
    # Change the configuration of the shell logging functionality
    shell_logging:
        # Always enable shell logging
        always_enable: {self.always_enable_shell_logging}
    
        #Choice of the method used to record the sessions (script or asciinema)
        logging_method: {self.shell_logging_method}
        
        # Enable automatic compression of log files (with gzip)
        enable_log_compression: {self.shell_logging_compress}
        
    # Configure your Exegol Desktop
    desktop:
        # Enables or not the desktop mode by default
        # If this attribute is set to True, then using the CLI --desktop option will be inverted and will DISABLE the feature
        enabled_by_default: {self.desktop_default_enable}
        
        # Default desktop protocol,can be "http", or "vnc" (additional protocols to come in the future, check online documentation for updates).
        default_protocol: {self.desktop_default_proto}
        
        # Desktop service is exposed on localhost by default. If set to true, services will be exposed on localhost (127.0.0.1) otherwise it will be exposed on 0.0.0.0. This setting can be overwritten with --desktop-config
        localhost_by_default: {self.desktop_default_localhost}
    
    # Configure your Exegol networks
    network:
    
        # Default network mode for any new container
        default_network: {self.network_default_mode}
        
        # Fallback bridge network mode
        # If the default network (host) mode cannot be used, a "bridge" network will automatically be selected as a replacement. (This mode can be "nat" or "docker")
        fallback_network: {self.network_fallback_mode}
        
        # Network range dedicated for exegol containers
        # Each new container using 'nat' network will have a dedicated smaller sub-network within this range (default to the last private /16 class B available)
        exegol_dedicated_range: {self.network_dedicated_range}
        
        # Exegol dedicated sub-network netmask.
        # By default, docker creates huge subnets, but exegol overrides this by using a much smaller subnet mask to optimize the use of network slots. (default to /28 with CIDR format)
        exegol_default_netmask: {self.network_default_netmask}

    # List of custom images from non-official private registry. (Enterprise feature only) More info here: https://docs.exegol.com/wrapper/configuration#custom-images
    custom_images:
    #  - docker.io/user/registry
"""
        return config

    @staticmethod
    def __default_resource_location(folder_name: str) -> Path:
        local_src = ConstantConfig.src_root_path_obj / folder_name
        if local_src.is_dir():
            # If exegol is clone from github, exegol submodule is accessible from root src
            return local_src
        else:
            # Default path for pip installation
            return ConstantConfig.exegol_config_path / folder_name

    def _process_data(self) -> None:
        # Volume section
        volumes_data = self._raw_data.get("volumes", {})
        # Catch existing but empty section
        if volumes_data is None:
            volumes_data = {}
        self.my_resources_path = self._load_config_path(volumes_data, 'my_resources_path', self.my_resources_path)
        self.private_volume_path = self._load_config_path(volumes_data, 'private_workspace_path', self.private_volume_path)
        self.exegol_resources_path = self._load_config_path(volumes_data, 'exegol_resources_path', self.exegol_resources_path)
        self.exegol_images_path = self._load_config_path(volumes_data, 'exegol_images_path', self.exegol_images_path)

        # Config section
        config_data = self._raw_data.get("config", {})
        # Catch existing but empty section
        if config_data is None:
            config_data = {}
        self.auto_check_updates = self._load_config_bool(config_data, 'auto_check_update', self.auto_check_updates)
        self.auto_remove_images = self._load_config_bool(config_data, 'auto_remove_image', self.auto_remove_images)
        self.auto_update_workspace_fs = self._load_config_bool(config_data, 'auto_update_workspace_fs', self.auto_update_workspace_fs)
        user_shell = self._load_config_str(config_data, 'default_start_shell', self.default_start_shell)
        # Manually check shell
        if len(user_shell.split(' ')) > 1:
            logger.warning(f"The configuration is incorrect! "
                           f"The user has configured the 'default_start_shell' parameter with the value '{user_shell}' "
                           f"which cannot be more than one word. The default value will be used instead: '{self.default_start_shell}'.")
        else:
            self.default_start_shell = user_shell

        self.enable_exegol_resources = self._load_config_bool(config_data, 'enable_exegol_resources', self.enable_exegol_resources)

        # Shell_logging section
        shell_logging_data = config_data.get("shell_logging", {})
        self.shell_logging_method = self._load_config_str(shell_logging_data, 'logging_method', self.shell_logging_method, choices=self.shell_logging_method_options)
        self.shell_logging_compress = self._load_config_bool(shell_logging_data, 'enable_log_compression', self.shell_logging_compress)
        self.always_enable_shell_logging = self._load_config_bool(shell_logging_data, 'always_enable', self.always_enable_shell_logging)

        # Desktop section
        desktop_data = config_data.get("desktop", {})
        self.desktop_default_enable = self._load_config_bool(desktop_data, 'enabled_by_default', self.desktop_default_enable)
        self.desktop_default_proto = self._load_config_str(desktop_data, 'default_protocol', self.desktop_default_proto, choices=self.desktop_available_proto)
        self.desktop_default_localhost = self._load_config_bool(desktop_data, 'localhost_by_default', self.desktop_default_localhost)

        # Network section
        network_data = config_data.get("network", {})
        self.network_default_mode = self._load_config_str(network_data, 'default_network', self.network_default_mode, choices=NetworkUtils.get_options())
        self.network_dedicated_range = self._load_config_str(network_data, 'exegol_dedicated_range')  # Dynamic default
        self.network_default_netmask = NetworkUtils.parse_netmask(self._load_config_str(network_data, 'exegol_default_netmask', str(self.network_default_netmask)), default=self.network_default_netmask)
        if SessionHandler().pro_feature_access():
            self.network_fallback_mode = self._load_config_str(network_data, 'fallback_network', self.network_fallback_mode, choices={'nat', 'docker'})
        else:
            self.network_fallback_mode = 'docker'

        # Enterprise features
        if SessionHandler().enterprise_feature_access():
            self.custom_images = self._load_config_list_str(config_data, "custom_images")

    def get_configs(self) -> List[str]:
        """User configs getter each options"""
        configs = [
            f"User config file: [magenta]{self._file_path}[/magenta]",
            f"Private workspace: [magenta]{self.private_volume_path}[/magenta]",
            "Exegol resources: " + (f"[magenta]{self.exegol_resources_path}[/magenta]"
                                    if self.enable_exegol_resources else
                                    boolFormatter(self.enable_exegol_resources)),
            f"Exegol images: [magenta]{self.exegol_images_path}[/magenta]",
            f"My resources: [magenta]{self.my_resources_path}[/magenta]",
            f"Auto-check updates: {boolFormatter(self.auto_check_updates)}",
            f"Auto-remove images: {boolFormatter(self.auto_remove_images)}",
            f"Auto-update fs: {boolFormatter(self.auto_update_workspace_fs)}",
            f"Default start shell: [blue]{self.default_start_shell}[/blue]",
            f"Always enable Shell logging: [blue]{boolFormatter(self.always_enable_shell_logging)}[/blue]",
            f"Shell logging method: [blue]{self.shell_logging_method}[/blue]",
            f"Shell logging compression: {boolFormatter(self.shell_logging_compress)}",
            f"Desktop enabled by default: {boolFormatter(self.desktop_default_enable)}",
            f"Desktop default protocol: [blue]{self.desktop_default_proto}[/blue]",
            f"Desktop default host: [blue]{'localhost' if self.desktop_default_localhost else '0.0.0.0'}[/blue]",
            f"Network default mode: [blue]{self.network_default_mode}[/blue]",
            f"Network fallback mode: [blue]{self.network_fallback_mode}[/blue]",
            f"Network range: [blue]{self.network_dedicated_range}[/blue]",
            f"Network exegol netmask: [blue]{self.network_default_netmask}[/blue]",
        ]
        if SessionHandler().enterprise_feature_access():
            if len(self.custom_images) > 0:
                configs.append(f"Custom images:")
                for custom_image in self.custom_images:
                    configs.append(f"  - {custom_image}")
            else:
                configs.append(f"Custom images: [bright_black]Empty[/bright_black]")
        # TUI can't be called from here to avoid circular importation
        return configs



import datetime
from typing import Optional

from pydantic import BaseModel


class SupabaseImage(BaseModel):
    """Supabase public Images Schema."""

    # Digests
    digest: str
    repo_digest: str

    # Columns
    tag: str
    arch: str
    build_date: datetime.datetime
    disk_size: float
    download_size: Optional[float]
    repository: str
    version: str
    license: Optional[str]

    def __repr__(self) -> str:
        return f"{self.repository}:{self.tag} {self.version} ({self.arch})"



import re
from datetime import datetime
from typing import Tuple, Union


# Generic text generation functions

def boolFormatter(val: bool) -> str:
    """Generic text formatter for bool value"""
    return '[green]On :heavy_check_mark:[/green] ' if val else '[orange3]Off :axe:[/orange3]'


def getColor(val: Union[bool, int, str]) -> Tuple[str, str]:
    """Generic text color getter for bool value"""
    if type(val) is str:
        try:
            val = int(val)
        except ValueError:
            val = False
    return ('[green]', '[/green]') if val else ('[orange3]', '[/orange3]')


def richLen(text: str) -> int:
    """Get real length of a text without Rich colors"""
    # remove rich color tags
    color_removed = re.sub(r"\[/?[^]]+]", '', text, 0, re.MULTILINE)
    # replace emoji by two random char (because emoji are wide)
    emoji_removed = re.sub(r":[a-z-_+()\d'’.&]+:", 'XX', color_removed, 0, re.MULTILINE)
    return len(emoji_removed)


def getArchColor(arch: str) -> str:
    if arch.startswith("arm"):
        color = "slate_blue3"
    elif "amd64" == arch:
        color = "medium_orchid3"
    else:
        color = "yellow3"
    return color


def get_display_date(date: Union[str, datetime]) -> str:
    if type(date) is str:
        if date == '':
            return ''
        date = parse_date(date)
    assert type(date) is datetime
    return date.astimezone().strftime("%d %B %Y %H:%M")


def parse_date(date: str) -> datetime:
    """Parse an ISO date string to a datetime object"""
    try:
        return datetime.fromisoformat(date)
    except ValueError:
        # Handle date parsing for python before 3.11
        from dateutil.parser import isoparse
        return isoparse(date)



import datetime
from typing import List, Optional, Dict, Union, Sequence, cast

from exegol.config.ConstantConfig import ConstantConfig
from exegol.utils.ExeLog import logger


class MetadataCacheModel:
    """MetadataCacheModel store a timestamp to compare with the last update"""

    def __init__(self, last_check=None, time_format: str = "%d/%m/%Y") -> None:
        self.__TIME_FORMAT = time_format
        if last_check is None:
            last_check = datetime.date.today().strftime(self.__TIME_FORMAT)
        self.last_check: str = last_check

    def update_last_check(self) -> None:
        self.last_check = datetime.date.today().strftime(self.__TIME_FORMAT)

    def get_last_check(self) -> datetime.datetime:
        return datetime.datetime.strptime(self.last_check, self.__TIME_FORMAT)

    def get_last_check_text(self) -> str:
        return self.last_check

    def is_outdated(self, days: int = 15, hours: int = 0) -> bool:
        """Check if the cache must be considered as expired."""
        now = datetime.datetime.now()
        last_check = self.get_last_check()
        if last_check > now:
            logger.debug("Incoherent last check date detected. Metadata must be updated.")
            return True
        # Check if the last update is older than the max delay configures (by default, return True after at least 15 days)
        return (last_check + datetime.timedelta(days=days, hours=hours)) < now


class ImageCacheModel:
    """This model store every important information for image caching."""

    def __init__(self, name: str, last_version: str, digest: str, source: str):
        self.name: str = name
        self.last_version: str = last_version
        self.digest: str = digest
        self.source: str = source

    def __str__(self) -> str:
        return f"{self.name}:{self.last_version} ({self.source}: {self.digest})"

    def __repr__(self) -> str:
        return str(self)


class ImagesCacheModel:
    """This model can store multiple image and the date of the last update"""

    def __init__(self, data: Sequence[Union[ImageCacheModel, Dict]], metadata: Optional[Dict] = None):
        # An old default date will be used until data are provided
        default_date: Optional[str] = "01/01/1990" if len(data) == 0 else None
        # Create or load (meta)data
        self.metadata: MetadataCacheModel = MetadataCacheModel(default_date) if metadata is None else MetadataCacheModel(**metadata)
        self.data: List[ImageCacheModel] = []
        if len(data) > 0:
            if type(data[0]) is dict:
                for img in data:
                    self.data.append(ImageCacheModel(**cast(Dict, img)))
            elif type(data[0]) is ImageCacheModel:
                self.data = cast(List[ImageCacheModel], data)
            else:
                raise NotImplementedError

    def __str__(self) -> str:
        return f"{len(self.data)} images ({self.metadata.last_check})"

    def __repr__(self) -> str:
        return str(self)


class WrapperCacheModel:
    """Caching wrapper update information (last version / last update)"""

    def __init__(self, last_version: Optional[str] = None, current_version: Optional[str] = None, metadata: Optional[Dict] = None):
        default_date: Optional[str] = None
        if last_version is None:
            last_version = ConstantConfig.version
            default_date = "01/01/1990"
        if current_version is None:
            current_version = ConstantConfig.version
        self.metadata: MetadataCacheModel = MetadataCacheModel(default_date) if metadata is None else MetadataCacheModel(**metadata)
        self.last_version: str = last_version
        self.current_version: str = current_version

    def __str__(self) -> str:
        return f"{self.current_version} -> {self.last_version} ({self.metadata.last_check})"

    def __repr__(self) -> str:
        return str(self)


class CacheDB:
    """Main object of the exegol cache"""

    def __init__(self) -> None:
        self.wrapper: WrapperCacheModel = WrapperCacheModel()
        self.images: ImagesCacheModel = ImagesCacheModel([])

    def load(self, wrapper: Dict, images: Dict) -> None:
        """Load the CacheDB data from a raw Dict object"""
        self.wrapper = WrapperCacheModel(**wrapper)
        self.images = ImagesCacheModel(**images)



import logging
import os
import re
import stat
import subprocess
import sys
from pathlib import Path, PurePath
from typing import Optional, Tuple

from exegol.config.EnvInfo import EnvInfo
from exegol.utils.ExeLog import logger


def parseDockerVolumePath(source: str) -> PurePath:
    """Parse docker volume path to find the corresponding host path."""
    # Check if path is from Windows Docker Desktop
    matches = re.match(r"^/run/desktop/mnt/host/([a-z])(/.*)$", source, re.IGNORECASE)
    if matches:
        # Convert Windows Docker-VM style volume path to local OS path
        src_path = Path(f"{matches.group(1).upper()}:{matches.group(2)}")
        logger.debug(f"Windows style detected : {src_path}")
        return src_path
    else:
        # Remove docker mount path if exist
        return PurePath(source.replace('/run/desktop/mnt/host', ''))


def resolvPath(path: Path) -> str:
    """Resolv a filesystem path depending on the environment.
    On WSL, Windows PATH can be resolved using 'wslpath'."""
    if path is None:
        return ''
    # From WSL, Windows Path must be resolved (try to detect a Windows path with '\')
    if EnvInfo.current_platform == "WSL" and '\\' in str(path):
        try:
            # Resolv Windows path on WSL environment
            p = subprocess.Popen(["wslpath", "-a", str(path)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output, err = p.communicate()
            logger.debug(f"Resolv path input: {path}")
            logger.debug(f"Resolv path output: {output!r}")
            if err != b'':
                # result is returned to STDERR when the translation didn't properly find a match
                logger.debug(f"Error on FS path resolution: {err!r}. Input path is probably a linux path.")
            else:
                return output.decode('utf-8').strip()
        except FileNotFoundError:
            logger.warning("Missing WSL tools: 'wslpath'. Skipping resolution.")
    return str(path)


def resolvStrPath(path: Optional[str]) -> str:
    """Try to resolv a filesystem path from a string."""
    if path is None:
        return ''
    return resolvPath(Path(path))


def setGidPermission(root_folder: Path) -> None:
    """Set the setgid permission bit to every recursive directory"""
    logger.verbose(f"Updating the permissions of {root_folder} (and sub-folders) to allow file sharing between the container and the host user")
    logger.debug(f"Adding setgid permission recursively on directories from {root_folder}")
    perm_alert = False
    # Set permission to root directory
    try:
        root_folder.chmod(root_folder.stat().st_mode | stat.S_IRWXG | stat.S_ISGID)
    except PermissionError:
        # Trigger the error only if the permission is not already set
        if not root_folder.stat().st_mode & stat.S_ISGID:
            logger.warning(f"The permission of this directory ({root_folder}) cannot be automatically changed.")
            perm_alert = True
    for sub_item in root_folder.rglob('*'):
        # Find every subdirectory
        try:
            if not sub_item.is_dir():
                continue
        except PermissionError:
            if not sub_item.is_symlink():
                logger.error(f"Permission denied when trying to resolv {str(sub_item)}")
            continue
        # If the permission is already set, skip
        if sub_item.stat().st_mode & stat.S_ISGID:
            continue
        # Set the permission (g+s) to every child directory
        try:
            sub_item.chmod(sub_item.stat().st_mode | stat.S_IRWXG | stat.S_ISGID)
        except PermissionError:
            logger.warning(f"The permission of this directory ({sub_item}) cannot be automatically changed.")
            perm_alert = True
    if perm_alert:
        logger.warning(f"In order to share files between your host and exegol (without changing the permission), you can run [orange3]manually[/orange3] this command from your [red]host[/red]:")
        logger.empty_line()
        logger.raw(f"sudo chgrp -R $(id -g) {root_folder} && sudo find {root_folder} -type d -exec chmod g+rws {{}} \\;", level=logging.WARNING)
        logger.empty_line()
        logger.empty_line()


def check_sysctl_value(sysctl: str, compare_to: str) -> bool:
    """Function to find a sysctl configured value and compare it to a desired value."""
    sysctl_path = "/proc/sys/" + sysctl.replace('.', '/')
    try:
        with open(sysctl_path, 'r') as conf:
            config = conf.read().strip()
            logger.debug(f"Checking sysctl value {sysctl}={config} (compare to {compare_to})")
            return conf.read().strip() == compare_to
    except FileNotFoundError:
        logger.debug(f"Sysctl file {sysctl} not found!")
    except PermissionError:
        logger.debug(f"Unable to read sysctl {sysctl} permission!")
    return False


def get_user_id() -> Tuple[int, int]:
    """On linux system, retrieve the original user id when using SUDO."""
    if sys.platform == "win32":
        raise SystemError
    user_uid_raw = os.getenv("SUDO_UID")
    if user_uid_raw is None:
        user_uid = os.getuid()
    else:
        user_uid = int(user_uid_raw)
    user_gid_raw = os.getenv("SUDO_GID")
    if user_gid_raw is None:
        user_gid = os.getgid()
    else:
        user_gid = int(user_gid_raw)
    return user_uid, user_gid


def mkdir(path: Path) -> None:
    """Function to recursively create a directory and setting the right user and group id to allow host user access."""
    if not path.parent.is_dir():
        mkdir(path.parent)
    try:
        path.mkdir(parents=False, exist_ok=False)
        if sys.platform == "linux" and os.getuid() == 0:
            user_uid, user_gid = get_user_id()
            os.chown(path, user_uid, user_gid)
    except FileExistsError:
        # The directory already exist, this setup can be skipped
        pass
    except (FileNotFoundError, PermissionError):
        logger.error(f"Unable to create directory {path}. Please check your file permissions.")
